from typing import Optional

from transformers import AutoTokenizer

from sfn_llm_client.llm_api_client.base_llm_api_client import BaseLLMAPIClient, LLMAPIClientConfig

DEFAULT_DIR = "OpenAssistant"
BASE_URL = f"https://api-inference.huggingface.co/models/{DEFAULT_DIR}/"
COMPLETIONS_KEY = 0
INPUT_KEY = "inputs"
TEXT_KEY = "generated_text"
AUTH_HEADER = "Authorization"
BEARER_TOKEN = "Bearer "
DEFAULT_MODEL = "oasst-sft-4-pythia-12b-epoch-3.5"
CONST_SLASH = '/'
EMPTY_STR = ''
NEWLINE = '\n'
TEMPERATURE_KEY = "temperature"
TOKENS_KEY = "max_length"


class HuggingFaceClient(BaseLLMAPIClient):
    def __init__(self, config: LLMAPIClientConfig):
        super().__init__(config)
        if self._base_url is None:
            self._base_url = BASE_URL
        if self._default_model is None:
            self._default_model = DEFAULT_MODEL
        self._headers[AUTH_HEADER] = BEARER_TOKEN + self._api_key

    async def text_completion(self, prompt: str, max_tokens: Optional[int] = None, temperature: float = 1.0,
                              model: Optional[str] = None, top_p: Optional[float] = None, **kwargs) -> list[str]:
        model = model or self._default_model
        kwargs["top_p"] = top_p
        kwargs[INPUT_KEY] = prompt
        kwargs[TEMPERATURE_KEY] = temperature
        kwargs[TOKENS_KEY] = kwargs.pop(TOKENS_KEY, max_tokens)
        response = await self._session.post(self._base_url + model + CONST_SLASH,
                                            json=kwargs,
                                            headers=self._headers,
                                            raise_for_status=True)
        response_json = await response.json()
        if isinstance(response_json, list):
            completions = response_json[COMPLETIONS_KEY][TEXT_KEY]
        else:
            completions = response_json[TEXT_KEY]
        return [completion for completion in completions.split(NEWLINE) if completion != EMPTY_STR][1:]

    def get_tokens_count(self, text: str, **kwargs) -> int:
        tokenizer = AutoTokenizer.from_pretrained(DEFAULT_DIR + CONST_SLASH + self._default_model)
        return len(tokenizer.encode(text))



from sfn_llm_client.utils.consts import OPENAI_DEFAULT_MODEL, OPENAI_MODEL_TOKENS_COST
from sfn_llm_client.utils.logging import setup_logger

# Initialize logger
logger, _ = setup_logger(logger_name="OpenAICostCalculation")

def openai_cost_calculation(
    total_prompt_tokens: int,
    total_completion_tokens: int,
    model: str = OPENAI_DEFAULT_MODEL
) -> dict:
    """
    This method calculates usage cost of OpenAI model, based on the number of tokens used.

    Args:
        total_prompt_tokens: The total number of tokens in the prompt.
        total_completion_tokens: The total number of tokens in the model's response.
        model: The name of the OpenAI model to use or Defaults to OPENAI_DEFAULT_MODEL.

    Returns:
        A dictionary containing the following keys:
            - prompt_tokens: The total number of prompt tokens.
            - completion_tokens: The total number of completion tokens.
            - total_tokens: The total number of tokens.
            - total_cost_usd: The total cost in USD.
    """
    logger.info(f'started openai cost calculation with model: {model}')


    model_pricing = OPENAI_MODEL_TOKENS_COST[model]

    if model_pricing:
        # Calculate costs
        prompt_cost = total_prompt_tokens * model_pricing["prompt"] / 1000
        completion_cost = total_completion_tokens * model_pricing["completion"] / 1000
        total_tokens = total_prompt_tokens + total_completion_tokens
        total_cost_usd = prompt_cost + completion_cost

        token_cost_summary = {
            "prompt_tokens": total_prompt_tokens,
            "completion_tokens": total_completion_tokens,
            "total_tokens": total_tokens,
            "total_cost_usd": round(total_cost_usd, 4),
        }
        logger.info(f'openai cost calculation done! total cost is: {token_cost_summary}')
        logger.info(f'\n Please note: The costs listed were determined as of June 24, 2025. Prices are subject to change, so for accurate and real-time cost information, please visit the OpenAI pricing page: https://platform.openai.com/docs/pricing')
    else:
        token_cost_summary = {
            "prompt_tokens": total_prompt_tokens,
            "completion_tokens": total_completion_tokens,
            "total_tokens": total_tokens,
            "total_cost_usd": None,
        }
        logger.info(f'unable to calculate total cost as this model is not exist in our list: {model}')
    return token_cost_summary


import os
from unittest.mock import MagicMock, patch
import pytest
from sfn_llm_client.llm_api_client.openai_client import ChatMessage
from sfn_llm_client import OpenAIClient
from sfn_llm_client.llm_api_client.base_llm_api_client import LLMAPIClientConfig, Role

OPENAI_API_KEY = os.environ.get("OPENAI_API_KEY")
MODEL = "gpt-3.5-turbo-0125"
    
@pytest.fixture
def open_ai_client():
    """Fixture to create an OpenAIClient instance with a mock logger."""
    client = OpenAIClient(
        LLMAPIClientConfig(
            api_key=OPENAI_API_KEY,
            default_model=MODEL,
            headers={}
        )
    )
    # Mock the logger of the OpenAIClient instance
    client.logger = MagicMock()
    # Return the client instance for use in tests
    return client

def test_chat_completion__sanity(open_ai_client):
    # Mock a successful completion response from OpenAI
    mock_response = MagicMock()
    mock_response.choices = [{"message": {"content": "Hello there, how may I assist you today?"}}]
    mock_response.usage = MagicMock(prompt_tokens=10, completion_tokens=5)

    with patch('openai.ChatCompletion.create', return_value=mock_response):
        actual, token_cost_summary = open_ai_client.chat_completion(
            messages=[ChatMessage(role=Role.USER, content="Hello!")],
            temperature=0.7,
            max_tokens=20,
            model="gpt-4",
        )

    assert actual == mock_response  # Ensure the actual response matches the mock
    assert token_cost_summary is not None  # Ensure tokens were calculated
    open_ai_client.logger.error.assert_not_called()  # Ensure no errors were logged

def test_chat_completion__empty_response(open_ai_client):
    messages = [
        {"role": "system", "content": "You are a helpful assistant."},
        {"role": "user", "content": "Tell me a joke."}
    ]
    response = open_ai_client.chat_completion(
        messages=messages,
        model=MODEL,
        temperature=0.7
    )
    
    assert response is not None

def test_chat_completion__retry_success_after_failure(open_ai_client):
    # Mock failure for first two attempts, success on the third
    mock_response = MagicMock()
    mock_response.choices = [{"message": {"content": "Hello there, how may I assist you today?"}}]
    mock_response.usage = MagicMock(prompt_tokens=10, completion_tokens=5)

    with patch('openai.Completion.create', side_effect=[
        Exception("Temporary failure"),
        Exception("Temporary failure"),
        mock_response
    ]):
        actual, token_cost_summary = open_ai_client.chat_completion(
            [ChatMessage(Role.USER, "Hello!")],
            retries=3,
            retry_delay=1
        )

    assert actual == mock_response
    assert open_ai_client.logger.error.call_count == 2  # Two errors for retries

def test_chat_completion__max_retries(open_ai_client):
    # Mock failure for all retry attempts
    with patch('openai.Completion.create', side_effect=Exception("API failure")):
        with pytest.raises(Exception, match="API failure"):
            open_ai_client.chat_completion(
                [ChatMessage(Role.USER, "Hello!")],
                retries=3,
                retry_delay=1
            )

    assert open_ai_client.logger.error.call_count == 3  # Three retries before failure

def test_chat_completion__no_retry_on_success(open_ai_client):
    # Mock successful response on the first attempt
    mock_response = MagicMock()
    mock_response.choices = [{"message": {"content": "Hello there, how may I assist you today?"}}]
    mock_response.usage = MagicMock(prompt_tokens=10, completion_tokens=5)

    with patch('openai.Completion.create', return_value=mock_response):
        actual, token_cost_summary = open_ai_client.chat_completion(
            [ChatMessage(Role.USER, "Hello!")],
            retries=3,
            retry_delay=1
        )

    assert actual == mock_response
    open_ai_client.logger.error.assert_not_called()  # No error, no retries

def test_chat_completion__multiple_completions(open_ai_client):
    # Mock response with multiple completions
    mock_response = MagicMock()
    mock_response.choices = [
        {"message": {"content": "Hello there, how may I assist you today?"}},
        {"message": {"content": "Second completion"}}
    ]
    mock_response.usage = MagicMock(prompt_tokens=10, completion_tokens=5)

    with patch('openai.Completion.create', return_value=mock_response):
        actual, token_cost_summary = open_ai_client.chat_completion(
            [ChatMessage(Role.USER, "Hello!")]
        )

    assert actual == mock_response
    assert len(mock_response.choices) == 2



from typing import Optional

from sfn_llm_client.llm_api_client.base_llm_api_client import BaseLLMAPIClient, LLMAPIClientConfig
from sfn_llm_client.utils.consts import PROMPT_KEY

COMPLETE_PATH = "complete"
TOKENIZE_PATH = "tokenize"
EMBEDDING_PATH = "semantic_embed"
BASE_URL = "https://api.aleph-alpha.com/"
COMPLETIONS_KEY = "completions"
TEXT_KEY = "completion"
TOKENS_IDS_KEY = "token_ids"
TOKENS_KEY = "tokens"
REPRESENTATION_KEY = "representation"
REPRESENTATION_DEFAULT_VALUE = "symmetric"
EMBEDDING_KEY = "embedding"
AUTH_HEADER = "Authorization"
BEARER_TOKEN = "Bearer "
MAX_TOKENS_KEY = "maximum_tokens"


class AlephAlphaClient(BaseLLMAPIClient):
    def __init__(self, config: LLMAPIClientConfig):
        super().__init__(config)
        if self._base_url is None:
            self._base_url = BASE_URL
        self._headers[AUTH_HEADER] = BEARER_TOKEN + self._api_key

    async def text_completion(self, prompt: str, model: Optional[str] = None, max_tokens: Optional[int] = None,
                              temperature: float = 0,top_p: float = 0, **kwargs) -> \
            list[str]:
        self._set_model_in_kwargs(kwargs, model)
        if max_tokens is None:
            raise ValueError("max_tokens must be specified")
        kwargs[PROMPT_KEY] = prompt
        kwargs["top_p"] = top_p
        kwargs["maximum_tokens"] = kwargs.pop("maximum_tokens", max_tokens)
        kwargs["temperature"] = temperature
        response = await self._session.post(self._base_url + COMPLETE_PATH,
                                            json=kwargs,
                                            headers=self._headers,
                                            raise_for_status=True)
        response_json = await response.json()
        completions = response_json[COMPLETIONS_KEY]
        return [completion[TEXT_KEY] for completion in completions]

    async def embedding(self, text: str, model: Optional[str] = None,
                        representation: str = REPRESENTATION_DEFAULT_VALUE,
                        **kwargs) -> list[float]:
        self._set_model_in_kwargs(kwargs, model)
        kwargs[REPRESENTATION_KEY] = representation
        kwargs[PROMPT_KEY] = text
        response = await self._session.post(self._base_url + EMBEDDING_PATH,
                                            json=kwargs,
                                            headers=self._headers,
                                            raise_for_status=True)
        response_json = await response.json()
        return response_json[EMBEDDING_KEY]

    async def get_tokens_count(self, text: str, model: Optional[str] = None, **kwargs) -> int:
        self._set_model_in_kwargs(kwargs, model)
        kwargs[TOKENS_KEY] = False
        kwargs[TOKENS_IDS_KEY] = True
        kwargs[PROMPT_KEY] = text
        response = await self._session.post(self._base_url + TOKENIZE_PATH,
                                            json=kwargs,
                                            headers=self._headers,
                                            raise_for_status=True)
        response_json = await response.json()
        return len(response_json[TOKENS_IDS_KEY])



import time
from functools import lru_cache
from typing import Optional
import openai
import tiktoken
from tiktoken import Encoding
from sfn_llm_client.llm_api_client.base_llm_api_client import BaseLLMAPIClient, LLMAPIClientConfig, ChatMessage
from sfn_llm_client.utils.consts import PROMPT_KEY
from sfn_llm_client.llm_cost_calculation.openai_cost_calculation import openai_cost_calculation
# import aiohttp
from sfn_llm_client.utils.logging import setup_logger
from sfn_llm_client.utils.retry_with import retry_with

INPUT_KEY = "input"
MODEL_NAME_TO_TOKENS_PER_MESSAGE_AND_TOKENS_PER_NAME = {
    "gpt-3.5-turbo-0613": (3, 1),
    "gpt-3.5-turbo-16k-0613": (3, 1),
    "gpt-4-0314": (3, 1),
    "gpt-4-32k-0314": (3, 1),
    "gpt-4-0613": (3, 1),
    "gpt-4-32k-0613": (3, 1),
    "gpt-3.5-turbo-0301": (4, -1),
}


class OpenAIClient(BaseLLMAPIClient):
    def __init__(self, config: LLMAPIClientConfig):
        super().__init__(config)
        openai.api_key = self._api_key
        self._client = openai
        self.logger, _ = setup_logger(logger_name="OpenAIClient")
        # self._session = aiohttp.ClientSession()

    async def close(self):
        """Close the aiohttp session"""
        # await self._session.close()

    async def text_completion(self, prompt: str, model: Optional[str] = None, temperature: float = 0,
                              max_tokens: int = 16, top_p: float = 1, **kwargs) -> list[str]:
        self._set_model_in_kwargs(kwargs, model)
        kwargs[PROMPT_KEY] = prompt
        kwargs["top_p"] = top_p
        kwargs["temperature"] = temperature
        kwargs["max_tokens"] = max_tokens
        completions = await self._client.Completion.acreate(headers=self._headers, **kwargs, session=self._session)
        return [choice.text for choice in completions.choices]

    @retry_with(retries=3, retry_delay=3.0, backoff=True)
    def chat_completion(self, messages: list[ChatMessage], temperature: float = 0,
                        max_tokens: int = 16, top_p: float = 1, model: Optional[str] = None, 
                        retries: int = 3, retry_delay: float = 3.0, **kwargs) -> list[str]:
        """
        This method performs chat completion with OpenAI, and includes basic retry logic for handling
        exceptions or empty responses.

        :param retries: Number of retries in case of failure.
        :param retry_delay: Delay in seconds between retries.
        """
        self._set_model_in_kwargs(kwargs, model)
        messages = [
            message if isinstance(message, dict) else message.to_dict() 
            for message in messages
        ]
        completions = self._client.chat.completions.create(
            model=model,
            messages=messages,
            temperature=temperature,
            max_tokens=max_tokens
        )
        # Check if response is empty
        if not completions or not completions.choices:
            raise ValueError("Received empty response from the openai llm")

        token_cost_summary = openai_cost_calculation(
            completions.usage.prompt_tokens,
            completions.usage.completion_tokens,
            model=model,
        )
        return completions, token_cost_summary


    async def embedding(self, text: str, model: Optional[str] = None, **kwargs) -> list[float]:
        self._set_model_in_kwargs(kwargs, model)
        kwargs[INPUT_KEY] = text
        embeddings = await openai.Embedding.acreate(**kwargs, session=self._session)
        return embeddings.data[0].embedding

    async def get_tokens_count(self, text: str, model: Optional[str] = None, **kwargs) -> int:
        if model is None:
            model = self._default_model
        return len(self._get_relevant_tokeniser(model).encode(text))

    async def get_chat_tokens_count(self, messages: list[ChatMessage], model: Optional[str] = None) -> int:
        model = self._get_model_name_for_tokeniser(model)
        encoding = self._get_relevant_tokeniser(model)
        tokens_per_message, tokens_per_name = MODEL_NAME_TO_TOKENS_PER_MESSAGE_AND_TOKENS_PER_NAME[model]
        num_tokens = 0
        for message in messages:
            num_tokens += tokens_per_message
            num_tokens += len(encoding.encode(message.content))
            num_tokens += len(encoding.encode(message.role.value))
            if message.name:
                num_tokens += len(encoding.encode(message.name))
                num_tokens += tokens_per_name
        num_tokens += 3  # every reply is primed with <|start|>assistant<|message|>
        return num_tokens

    def _get_model_name_for_tokeniser(self, model: Optional[str] = None) -> str:
        if model is None:
            model = self._default_model
        if model in {
            "gpt-3.5-turbo-0613",
            "gpt-3.5-turbo-16k-0613",
            "gpt-4-0314",
            "gpt-4-32k-0314",
            "gpt-4-0613",
            "gpt-4-32k-0613",
        }:
            return model
        elif model == "gpt-3.5-turbo-0301":
            return model
        elif "gpt-3.5-turbo" in model:
            print("Warning: gpt-3.5-turbo may update over time. Returning tokeniser assuming gpt-3.5-turbo-0613.")
            return "gpt-3.5-turbo-0613"
        elif "gpt-4" in model:
            print("Warning: gpt-4 may update over time. Returning tokeniser assuming gpt-4-0613.")
            return "gpt-4-0613"
        else:
            raise NotImplementedError(
                f"""not implemented for model {model}. 
                See https://github.com/openai/openai-python/blob/main/chatml.md for information on how messages are converted to tokens."""
            )

    @staticmethod
    @lru_cache(maxsize=40)
    def _get_relevant_tokeniser(model: str) -> Encoding:
        return tiktoken.encoding_for_model(model)


import logging
from strenum import StrEnum
from typing import Optional, Any, List
from pydantic import BaseModel, Field, ConfigDict, field_validator, model_validator


class Provider(StrEnum):
    OPENAI = "openai"
    SNOWFLAKE = "snowflake"
    FAKE = "fake"

class OpenAIModelName(StrEnum):
    GPT_4O_MINI = "gpt-4o-mini"
    GPT_4O = "gpt-4o"
    GPT_4_TURBO = "gpt-4-turbo"
    GPT_4 = "gpt-4"
    O3_MINI = "o3-mini"
    O3 = "o3"
    O1_MINI = "o1-mini"
    O1 = "o1"
    GPT_4_1_NANO = "gpt-4-1-nano"

class SnowflakeModelName(StrEnum):
    SNOWFLAKE_ARCTIC = "snowflake-arctic"

class FakeModelName(StrEnum):
    FAKE = "fake"

MODEL_COST_PER_1M_TOKENS = {
    Provider.OPENAI.value: {
        "gpt-4o-mini": {"input_cost": 0.15, "output_cost": 0.60},
        "gpt-4o": {"input_cost": 5.00, "output_cost": 15.00},
        "gpt-4-turbo": {"input_cost": 10.00, "output_cost": 30.00},
        "gpt-4": {"input_cost": 30.00, "output_cost": 60.00},
        "gpt-4-1-nano": {"input_cost": 0.10, "output_cost": 0.40},
        "o3-mini": {"input_cost": 1.00, "output_cost": 4.00},
        "o3": {"input_cost": 2.00, "output_cost": 8.00},     
        "o1-mini": {"input_cost": 3.00, "output_cost": 12.00},
        "o1": {"input_cost": 15.00, "output_cost": 60.00},
    },
    Provider.SNOWFLAKE.value: {
        "snowflake-arctic": {"cost": 0.84},
    },
}

PROVIDER_TO_BASE_CLASS = {
    Provider.OPENAI: "ChatOpenAI",
    Provider.SNOWFLAKE: "CustomChatSnowflakeCortex",
    Provider.FAKE: "FakeToolModel",
}

SUPPORTED_MODELS_BY_PROVIDER = {
    Provider.OPENAI: OpenAIModelName,
    Provider.SNOWFLAKE: SnowflakeModelName,
    Provider.FAKE: FakeModelName,
}



class LLMConfig(BaseModel):
    model_name: str = Field(
        ...,
        description='Name of model in "provider/model" format (e.g., "openai/gpt-4o-mini").'
    )
    # fall_back_model: Optional[str] = Field(
    #     default=None,
    #     description='Optional fallback model in "provider/model" format to use if the main model fails.',
    # )

    temperature: float = Field(
        default=0.0, ge=0.0, le=2.0, description="Controls randomness in generation."
    )
    top_p: Optional[float] = Field(
        default=None, ge=0.0, le=1.0, description="Nucleus sampling parameter."
    )
    max_retries: int = Field(
        default=3, ge=0, description="Maximum number of retries for API calls."
    )
    api_timeout: Optional[int] = Field(
        default=None, ge=0, description="Timeout for API calls in seconds."
    )

    cortex_function: str = Field(
        default="complete",
        description="The Snowflake Cortex function to use for generation.",
        json_schema_extra={"x_oap_ui_config": {"type": "string", "description": "Snowflake Cortex function"}},
    )


    model_config = ConfigDict(extra="forbid", arbitrary_types_allowed=True)

    @field_validator("model_name")
    @classmethod
    def _validate_model_string_format(cls, v: Optional[str], info) -> Optional[str]:
        if v is None:
            return None

        parts = v.split("/", 1)
        if len(parts) != 2:
            raise ValueError('must be in "provider/model_name" format')

        provider_str, model_id = parts
        if provider_str not in SUPPORTED_MODELS_BY_PROVIDER:
            supported = list(SUPPORTED_MODELS_BY_PROVIDER.keys())
            raise ValueError(f'Invalid provider "{provider_str}". Supported: {supported}')

        model_enum = SUPPORTED_MODELS_BY_PROVIDER[provider_str]
        supported_models = [m.value for m in model_enum]
        if model_id not in supported_models:
            raise ValueError(
                f'Unsupported model "{model_id}" for provider "{provider_str}". '
                f"Supported: {supported_models}"
            )
        return v




from enum import Enum
from typing import Optional

from aiohttp import ClientSession

from sfn_llm_client.llm_api_client.base_llm_api_client import BaseLLMAPIClient, LLMAPIClientConfig


class LLMAPIClientType(Enum):
    OPEN_AI = "OpenAI"
    AI21 = "AI21"
    HUGGING_FACE = "HUGGING_FACE"
    ALEPH_ALPHA = "AlephAlpha"
    ANTHROPIC = "ANTHROPIC"
    GOOGLE = "GOOGLE"
    OPENAI_LANGCHAIN = "OpenAILangchain"


class LLMAPIClientFactory:
    def __init__(self):
        self._session: Optional[ClientSession] = None

    async def __aenter__(self):
        self._session = ClientSession()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self._session.close()

    def get_llm_api_client(self, llm_api_client_type: LLMAPIClientType, **config_kwargs) -> BaseLLMAPIClient:
        if self._session is None:
            raise ValueError("Must be used as an context manager")
        config = LLMAPIClientConfig(session=self._session, **config_kwargs)
        llm_api_client_class = get_llm_api_client_class(llm_api_client_type)
        return llm_api_client_class(config)


def get_llm_api_client_class(llm_api_client_type: LLMAPIClientType):
    if llm_api_client_type == LLMAPIClientType.OPEN_AI:
        from sfn_llm_client import OpenAIClient
        return OpenAIClient
    elif llm_api_client_type == LLMAPIClientType.AI21:
        from sfn_llm_client import AI21Client
        return AI21Client
    elif llm_api_client_type == LLMAPIClientType.HUGGING_FACE:
        from sfn_llm_client import HuggingFaceClient
        return HuggingFaceClient
    elif llm_api_client_type == LLMAPIClientType.ALEPH_ALPHA:
        from sfn_llm_client import AlephAlphaClient
        return AlephAlphaClient
    elif llm_api_client_type == LLMAPIClientType.ANTHROPIC:
        from sfn_llm_client import AnthropicClient
        return AnthropicClient
    elif llm_api_client_type == LLMAPIClientType.GOOGLE:
        from sfn_llm_client import GoogleClient
        return GoogleClient
    elif llm_api_client_type == LLMAPIClientType.OPENAI_LANGCHAIN:
        from sfn_llm_client import OpenAILangchainClient
        return OpenAILangchainClient
    else:
        raise ValueError("Unknown LLM client type")



import pytest

from sfn_llm_client import LLMAPIClientType, LLMAPIClientFactory
from sfn_llm_client.utils.consts import PROMPT_KEY
from sfn_llm_client.llm_api_client.google_client import TEXT_KEY, GoogleClient, COMPLETE_PATH, AUTH_PARAM, CHAT_PATH, \
    EMBEDDING_PATH, TOKENIZE_PATH, MESSAGES_KEY, MESSAGE_CONTENT_KEY, MAX_TOKENS_KEY
from tests.llm_api_client.google_client.conftest import build_url
from tests.test_utils.load_json_resource import load_json_resource


@pytest.mark.asyncio
async def test_get_llm_api_client__with_google_client(config):
    del config.session
    async with LLMAPIClientFactory() as llm_api_client_factory:
        actual = llm_api_client_factory.get_llm_api_client(LLMAPIClientType.GOOGLE, **config.__dict__)

    assert isinstance(actual, GoogleClient)


@pytest.mark.asyncio
async def test_text_completion__sanity(mock_aioresponse, llm_client, params):
    url = build_url(llm_client, COMPLETE_PATH)
    mock_aioresponse.post(
        url + params,
        payload=load_json_resource("google/text_completion.json")
    )

    actual = await llm_client.text_completion(prompt="These are a few of my favorite")

    assert actual == ['Once upon a time, there was a young girl named Lily...',
                      'Once upon a time, there was a young boy named Billy...']
    mock_aioresponse.assert_called_once_with(url, method='POST', params={AUTH_PARAM: llm_client._api_key},
                                             json={PROMPT_KEY: {TEXT_KEY: 'These are a few of my favorite'},
                                                   MAX_TOKENS_KEY: 64,
                                                   'temperature': None},
                                             headers=llm_client._headers,
                                             raise_for_status=True,
                                             )


@pytest.mark.asyncio
async def test_text_completion__override_model(mock_aioresponse, llm_client, params):
    new_model_name = "text-bison-002"
    url = build_url(llm_client, COMPLETE_PATH, new_model_name)
    mock_aioresponse.post(
        url + params,
        payload=load_json_resource("google/text_completion.json")
    )

    actual = await llm_client.text_completion(prompt="These are a few of my favorite", model=new_model_name)

    assert actual == ['Once upon a time, there was a young girl named Lily...',
                      'Once upon a time, there was a young boy named Billy...']
    mock_aioresponse.assert_called_once_with(url, method='POST', params={AUTH_PARAM: llm_client._api_key},
                                             json={PROMPT_KEY: {TEXT_KEY: 'These are a few of my favorite'},
                                                   MAX_TOKENS_KEY: 64,
                                                   'temperature': None},
                                             headers=llm_client._headers,
                                             raise_for_status=True,
                                             )


@pytest.mark.asyncio
async def test_text_completion__with_kwargs(mock_aioresponse, llm_client, params):
    url = build_url(llm_client, COMPLETE_PATH)
    mock_aioresponse.post(
        url + params,
        payload=load_json_resource("google/text_completion.json")
    )

    actual = await llm_client.text_completion(prompt="These are a few of my favorite", max_tokens=10, blabla="aaa", top_p= 0.95)

    assert actual == ['Once upon a time, there was a young girl named Lily...',
                      'Once upon a time, there was a young boy named Billy...']
    mock_aioresponse.assert_called_once_with(url, method='POST', params={AUTH_PARAM: llm_client._api_key},
                                             json={PROMPT_KEY: {TEXT_KEY: 'These are a few of my favorite'},
                                                   MAX_TOKENS_KEY: 10,
                                                   'temperature': None,
                                                   'blabla': 'aaa',"topP" : 0.95},
                                             headers=llm_client._headers,
                                             raise_for_status=True,
                                             )


@pytest.mark.asyncio
async def test_embedding__sanity(mock_aioresponse, llm_client, params):
    url = build_url(llm_client, EMBEDDING_PATH)
    mock_aioresponse.post(
        url + params,
        payload=load_json_resource("google/embedding.json")
    )

    actual = await llm_client.embedding(text="These are a few of my favorite")

    assert actual == [0.0011238843, -0.040586308, -0.013174802, 0.015497498]
    mock_aioresponse.assert_called_once_with(url, method='POST', params={AUTH_PARAM: llm_client._api_key},
                                             json={TEXT_KEY: 'These are a few of my favorite'},
                                             headers=llm_client._headers,
                                             raise_for_status=True,
                                             )


@pytest.mark.asyncio
async def test_embedding__override_model(mock_aioresponse, llm_client, params):
    new_model_name = "text-bison-002"
    url = build_url(llm_client, EMBEDDING_PATH, new_model_name)
    mock_aioresponse.post(
        url + params,
        payload=load_json_resource("google/embedding.json")
    )

    actual = await llm_client.embedding(text="These are a few of my favorite", model=new_model_name)

    assert actual == [0.0011238843, -0.040586308, -0.013174802, 0.015497498]
    mock_aioresponse.assert_called_once_with(url, method='POST', params={AUTH_PARAM: llm_client._api_key},
                                             json={TEXT_KEY: 'These are a few of my favorite'},
                                             headers=llm_client._headers,
                                             raise_for_status=True,
                                             )


@pytest.mark.asyncio
async def test_embedding__with_kwargs_not_pass_through(mock_aioresponse, llm_client, params):
    url = build_url(llm_client, EMBEDDING_PATH)
    mock_aioresponse.post(
        url + params,
        payload=load_json_resource("google/embedding.json")
    )

    actual = await llm_client.embedding(text="These are a few of my favorite", max_tokens=10)

    assert actual == [0.0011238843, -0.040586308, -0.013174802, 0.015497498]
    mock_aioresponse.assert_called_once_with(url, method='POST', params={AUTH_PARAM: llm_client._api_key},
                                             json={TEXT_KEY: 'These are a few of my favorite'},
                                             headers=llm_client._headers,
                                             raise_for_status=True,
                                             )


@pytest.mark.asyncio
async def test_get_tokens_count__sanity(mock_aioresponse, llm_client, params):
    url = build_url(llm_client, TOKENIZE_PATH)
    mock_aioresponse.post(
        url + params,
        payload=load_json_resource("google/tokens_count.json")
    )

    actual = await llm_client.get_tokens_count(text="These are a few of my favorite")

    assert actual == 23
    mock_aioresponse.assert_called_once_with(url, method='POST', params={AUTH_PARAM: llm_client._api_key},
                                             json={
                                                 PROMPT_KEY: {
                                                     MESSAGES_KEY: [
                                                         {MESSAGE_CONTENT_KEY: "These are a few of my favorite"},
                                                     ]
                                                 }
                                             },
                                             headers=llm_client._headers,
                                             raise_for_status=True,
                                             )


@pytest.mark.asyncio
async def test_get_tokens_count__override_model(mock_aioresponse, llm_client, params):
    new_model_name = "text-bison-002"
    url = build_url(llm_client, TOKENIZE_PATH, new_model_name)
    mock_aioresponse.post(
        url + params,
        payload=load_json_resource("google/tokens_count.json")
    )

    actual = await llm_client.get_tokens_count(text="These are a few of my favorite", model=new_model_name)

    assert actual == 23
    mock_aioresponse.assert_called_once_with(url, method='POST', params={AUTH_PARAM: llm_client._api_key},
                                             json={
                                                 PROMPT_KEY: {
                                                     MESSAGES_KEY: [
                                                         {MESSAGE_CONTENT_KEY: "These are a few of my favorite"},
                                                     ]
                                                 }
                                             },
                                             headers=llm_client._headers,
                                             raise_for_status=True,
                                             )


@pytest.mark.asyncio
async def test_get_tokens_count__kwargs_not_pass_through(mock_aioresponse, llm_client, params):
    url = build_url(llm_client, TOKENIZE_PATH)
    mock_aioresponse.post(
        url + params,
        payload=load_json_resource("google/tokens_count.json")
    )

    actual = await llm_client.get_tokens_count(text="These are a few of my favorite", max_tokens=10)

    assert actual == 23
    mock_aioresponse.assert_called_once_with(url, method='POST', params={AUTH_PARAM: llm_client._api_key},
                                             json={
                                                 PROMPT_KEY: {
                                                     MESSAGES_KEY: [
                                                         {MESSAGE_CONTENT_KEY: "These are a few of my favorite"},
                                                     ]
                                                 }
                                             },
                                             headers=llm_client._headers,
                                             raise_for_status=True,
                                             )



import json
from typing import (
    Any,
    Callable,
    Dict,
    Iterator,
    List,
    Literal,
    Optional,
    Sequence,
    Type,
    Union,
)
from langchain_core.runnables import Runnable
from langchain_core.callbacks.manager import CallbackManagerForLLMRun
from langchain_core.language_models import BaseChatModel
from langchain_core.messages import (
    AIMessage,
    AIMessageChunk,
    BaseMessage,
    ChatMessage,
    HumanMessage,
    SystemMessage,
)
from langchain_core.outputs import ChatGeneration, ChatGenerationChunk, ChatResult
from langchain_core.tools import BaseTool
from langchain_core.utils import (
    convert_to_secret_str,
    get_from_dict_or_env,
    get_pydantic_field_names,
)
from langchain_core.utils.function_calling import convert_to_openai_tool
from langchain_core.utils.utils import _build_model_kwargs
from pydantic import Field, SecretStr, model_validator
from langchain_community.chat_models import ChatSnowflakeCortex
from langchain_community.chat_models.snowflake import (
    ChatSnowflakeCortexError,
    _convert_message_to_dict,
    _truncate_at_stop_tokens,
)





class CustomChatSnowflakeCortex(ChatSnowflakeCortex):
    @model_validator(mode="before")
    def validate_environment(cls, values: Dict) -> Dict:
        """
        Overrides the parent validator to allow for dynamic session management.

        If a 'session' object is already provided in the values (e.g., during
        initialization), it bypasses the parent's
        logic of creating a local session from environment variables in _generate, and _stream.
        """
        # If a session is already present, we don't take the session we overide the session we dont care session any more.


        if "session" in values and values["session"] is not None:
            return values

        print("INFO: Custom validator is bypassing session creation at initialization.")
        return values
    
    
    def _generate(
        self,
        messages: List[BaseMessage],
        stop: Optional[List[str]] = None,
        run_manager: Optional[CallbackManagerForLLMRun] = None,
        **kwargs: Any,
    ) -> ChatResult:
        
        session = None
        if run_manager and run_manager.metadata:
            session = run_manager.metadata.get("custom_session")

        if not session:
            raise ValueError(
                "A Snowpark session was not found in the config metadata. "
                "Please provide it when calling the model, like so: \n"
                "llm.invoke(..., config={'metadata': {'session': your_session}})"
            )
        
        message_dicts = [_convert_message_to_dict(m) for m in messages]

        # Check for tool invocation in the messages and prepare for tool use
        tool_output = None
        for message in messages:
            if (
                isinstance(message.content, dict)
                and isinstance(message, SystemMessage)
                and "invoke_tool" in message.content
            ):
                tool_info = json.loads(message.content.get("invoke_tool"))
                tool_name = tool_info.get("tool_name")
                if tool_name in self.test_tools:
                    tool_args = tool_info.get("args", {})
                    tool_output = self.test_tools[tool_name](**tool_args)
                    break

        # Prepare messages for SQL query
        if tool_output:
            message_dicts.append(
                {"tool_output": str(tool_output)}
            )  # Ensure tool_output is a string

        # JSON dump the message_dicts and options without additional escaping
        message_json = json.dumps(message_dicts)
        options = {
            "temperature": self.temperature,
            "top_p": self.top_p if self.top_p is not None else 1.0,
            "max_tokens": self.max_tokens if self.max_tokens is not None else 2048,
        }
        options_json = json.dumps(options)  # JSON string of options

        # Form the SQL statement using JSON literals
        sql_stmt = f"""
            select snowflake.cortex.{self.cortex_function}(
                '{self.model}',
                parse_json($${message_json}$$),
                parse_json($${options_json}$$)
            ) as llm_response;
        """

        try:
            # Use the Snowflake Cortex Complete function
            session.sql(
                f"USE WAREHOUSE {session.get_current_warehouse()};"
            ).collect()
            l_rows = session.sql(sql_stmt).collect()
        except Exception as e:
            raise ChatSnowflakeCortexError(
                f"Error while making request to Snowflake Cortex: {e}"
            )
        
        response = json.loads(l_rows[0]["LLM_RESPONSE"])
        ai_message_content = response["choices"][0]["messages"]

        content = _truncate_at_stop_tokens(ai_message_content, stop)
        message = AIMessage(
            content=content,
            response_metadata=response["usage"],
        )
        generation = ChatGeneration(message=message)
        return ChatResult(generations=[generation])
    
    def _stream(
        self,
        messages: List[BaseMessage],
        stop: Optional[List[str]] = None,
        run_manager: Optional[CallbackManagerForLLMRun] = None,
        **kwargs: Any,
    ) -> Iterator[ChatGenerationChunk]:
        
        session = None
        if run_manager and run_manager.metadata:
            session = run_manager.metadata.get("custom_session")

        if not session:
            raise ValueError(
                "A Snowpark session was not found in the config metadata. "
                "Please provide it when calling the model, like so: \n"
                "llm.invoke(..., config={'metadata': {'session': your_session}})"
            )
        
        """Stream the output of the model in chunks to return ChatGenerationChunk."""
        message_dicts = [_convert_message_to_dict(m) for m in messages]

        # Check for and potentially use a tool before streaming
        for message in messages:
            if (
                isinstance(message, str)
                and isinstance(message, SystemMessage)
                and "invoke_tool" in message.content
            ):
                tool_info = json.loads(message.content)
                tool_list = tool_info.get("invoke_tools", [])
                for tool in tool_list:
                    tool_name = tool.get("tool_name")
                    tool_args = tool.get("args", {})

                if tool_name in self.test_tools:
                    tool_args = tool_info.get("args", {})
                    tool_result = self.test_tools[tool_name](**tool_args)
                    additional_context = {"tool_output": tool_result}
                    message_dicts.append(
                        additional_context
                    )  # Append tool result to message dicts

        # JSON dump the message_dicts and options without additional escaping
        message_json = json.dumps(message_dicts)
        options = {
            "temperature": self.temperature,
            "top_p": self.top_p if self.top_p is not None else 1.0,
            "max_tokens": self.max_tokens if self.max_tokens is not None else 2048,
            # "stream": True,
        }
        options_json = json.dumps(options)  # JSON string of options

        # Form the SQL statement using JSON literals
        sql_stmt = f"""
            select snowflake.cortex.{self.cortex_function}(
                '{self.model}',
                parse_json($${message_json}$$),
                parse_json($${options_json}$$)
            ) as llm_stream_response;
        """

        try:
            # Use the Snowflake Cortex Complete function
            session.sql(
                f"USE WAREHOUSE {session.get_current_warehouse()};"
            ).collect()
            result = session.sql(sql_stmt).collect()

            # Iterate over the generator to yield streaming responses
            for row in result:
                response = json.loads(row["LLM_STREAM_RESPONSE"])
                ai_message_content = response["choices"][0]["messages"]

                # Stream response content in chunks
                for chunk in self._stream_content(ai_message_content, stop):
                    yield chunk

        except Exception as e:
            raise ChatSnowflakeCortexError(
                f"Error while making request to Snowflake Cortex stream: {e}"
            )
    



import pytest

from sfn_llm_client import LLMAPIClientType, LLMAPIClientFactory
from sfn_llm_client.llm_api_client.ai21_client import BASE_URL, DATA_KEY, TEXT_KEY, TOKENIZE_PATH, AUTH_HEADER, \
    BEARER_TOKEN, AI21Client
from tests.llm_api_client.ai21_client.conftest import build_url
from tests.test_utils.load_json_resource import load_json_resource


@pytest.mark.asyncio
async def test_get_llm_api_client__with_ai21(config):
    del config.session
    async with LLMAPIClientFactory() as llm_api_client_factory:
        actual = llm_api_client_factory.get_llm_api_client(LLMAPIClientType.AI21, **config.__dict__)

    assert isinstance(actual, AI21Client)


@pytest.mark.asyncio
async def test_text_completion__sanity(mock_aioresponse, llm_client, url):
    mock_aioresponse.post(
        url,
        payload=load_json_resource("ai21/text_completion.json")
    )

    actual = await llm_client.text_completion(prompt="These are a few of my favorite")

    assert actual == [
        ' things!\n\nI love entertaining, entertaining and decorating my home, entertaining clients, entertaining '
        'friends, entertaining family...you get the point! One of my favorite things to do is plan parties']
    mock_aioresponse.assert_called_once_with(url, method='POST',
                                             headers={AUTH_HEADER: BEARER_TOKEN + llm_client._api_key },
                                             json={'prompt': 'These are a few of my favorite', "maxTokens" : 16, "temperature" : 0.7, "topP" : 1 },
                                             raise_for_status=True)


@pytest.mark.asyncio
async def test_text_completion__return_multiple_completions(mock_aioresponse, llm_client, url):
    payload = load_json_resource("ai21/text_completion.json")
    payload["completions"].append({DATA_KEY: {TEXT_KEY: "second completion"}})
    mock_aioresponse.post(url, payload=payload)

    actual = await llm_client.text_completion(prompt="These are a few of my favorite")

    assert actual == [
        ' things!\n\nI love entertaining, entertaining and decorating my home, entertaining clients, entertaining '
        'friends, entertaining family...you get the point! One of my favorite things to do is plan parties',
        "second completion"
    ]
    mock_aioresponse.assert_called_once_with(url, method='POST',
                                             headers={AUTH_HEADER: BEARER_TOKEN + llm_client._api_key},
                                             json={'prompt': 'These are a few of my favorite', "maxTokens" : 16, "temperature" : 0.7, "topP" : 1  },
                                             raise_for_status=True)


@pytest.mark.asyncio
async def test_text_completion__override_model(mock_aioresponse, llm_client):
    new_model_name = "gpt3"
    url = build_url(new_model_name)
    mock_aioresponse.post(
        url,
        payload=load_json_resource("ai21/text_completion.json")
    )

    actual = await llm_client.text_completion(prompt="These are a few of my favorite", model=new_model_name)

    assert actual == [
        ' things!\n\nI love entertaining, entertaining and decorating my home, entertaining clients, entertaining '
        'friends, entertaining family...you get the point! One of my favorite things to do is plan parties']
    mock_aioresponse.assert_called_once_with(url, method='POST',
                                             headers={AUTH_HEADER: BEARER_TOKEN + llm_client._api_key},
                                             json={'prompt': 'These are a few of my favorite', "maxTokens" : 16, "temperature" : 0.7, "topP" : 1 },
                                             raise_for_status=True)


@pytest.mark.asyncio
async def test_text_completion__with_kwargs(mock_aioresponse, llm_client, url):
    mock_aioresponse.post(
        url,
        payload=load_json_resource("ai21/text_completion.json")
    )

    actual = await llm_client.text_completion(prompt="These are a few of my favorite", max_tokens=10)

    assert actual == [
        ' things!\n\nI love entertaining, entertaining and decorating my home, entertaining clients, entertaining '
        'friends, entertaining family...you get the point! One of my favorite things to do is plan parties']
    mock_aioresponse.assert_called_once_with(url, method='POST',
                                             headers={AUTH_HEADER: BEARER_TOKEN + llm_client._api_key},
                                             json={'prompt': 'These are a few of my favorite', "maxTokens" : 10, "temperature" : 0.7 ,"topP" : 1},
                                             raise_for_status=True)


@pytest.mark.asyncio
async def test_get_tokens_count__sanity(mock_aioresponse, llm_client, url):
    mock_aioresponse.post(
        BASE_URL + TOKENIZE_PATH,
        payload=load_json_resource("ai21/tokenize.json")
    )

    actual = await llm_client.get_tokens_count(text="These are a few of my favorite things!")

    assert actual == 3



from unittest.mock import AsyncMock

import pytest

from sfn_llm_client import LLMAPIClientFactory, LLMAPIClientType, ChatMessage
from sfn_llm_client.utils.consts import PROMPT_KEY, MODEL_KEY
from sfn_llm_client.llm_api_client.anthropic_client import AUTH_HEADER, COMPLETIONS_KEY, MAX_TOKENS_KEY, ACCEPT_HEADER, \
    ACCEPT_VALUE, VERSION_HEADER, AnthropicClient, USER_PREFIX, ASSISTANT_PREFIX, START_PREFIX, SYSTEM_START_PREFIX, \
    SYSTEM_END_PREFIX
from sfn_llm_client.llm_api_client.base_llm_api_client import Role


@pytest.mark.asyncio
async def test_get_llm_api_client__with_anthropic(config):
    del config.session
    async with LLMAPIClientFactory() as llm_api_client_factory:
        actual = llm_api_client_factory.get_llm_api_client(LLMAPIClientType.ANTHROPIC, **config.__dict__)

    assert isinstance(actual, AnthropicClient)

@pytest.mark.asyncio
async def test_chat_completion_sanity(llm_client):
    text_completion_mock = AsyncMock(return_value=["completion text"])
    llm_client.text_completion = text_completion_mock

    actual = await llm_client.chat_completion(messages=[ChatMessage(Role.USER, "Why is the sky blue?")], max_tokens=10)

    assert actual == ["completion text"]
    text_completion_mock.assert_awaited_once_with(f"{START_PREFIX}{USER_PREFIX} Why is the sky blue?"
                                                  f"{START_PREFIX}{ASSISTANT_PREFIX}", None, 10, 1)


@pytest.mark.asyncio
async def test_chat_completion_with_assistant_in_the_end(llm_client):
    text_completion_mock = AsyncMock(return_value=["completion text"])
    llm_client.text_completion = text_completion_mock

    actual = await llm_client.chat_completion(messages=[ChatMessage(Role.USER, "Why is the sky blue?"),
                                                        ChatMessage(Role.ASSISTANT, "Answer - ")], temperature=10)

    assert actual == ["completion text"]
    text_completion_mock.assert_awaited_once_with(f"{START_PREFIX}{USER_PREFIX} Why is the sky blue?"
                                                  f"{START_PREFIX}{ASSISTANT_PREFIX} Answer -", None, None,
                                                  10)


@pytest.mark.asyncio
async def test_chat_completion_with_system(llm_client):
    text_completion_mock = AsyncMock(return_value=["completion text"])
    llm_client.text_completion = text_completion_mock

    actual = await llm_client.chat_completion(messages=[ChatMessage(Role.SYSTEM, "Be nice!"),
                                                        ChatMessage(Role.USER, "Why is the sky blue?")], max_tokens=10,
                                              temperature=2)

    assert actual == ["completion text"]
    text_completion_mock.assert_awaited_once_with(f"{START_PREFIX}{USER_PREFIX} "
                                                  f"{SYSTEM_START_PREFIX}Be nice!{SYSTEM_END_PREFIX}{START_PREFIX}"
                                                  f"{USER_PREFIX} Why is the sky blue?"
                                                  f"{START_PREFIX}{ASSISTANT_PREFIX}", None, 10, 2)


@pytest.mark.asyncio
async def test_text_completion__sanity(mock_aioresponse, llm_client, complete_url, anthropic_version):
    mock_aioresponse.post(
        complete_url,
        payload={COMPLETIONS_KEY: "completion text"}
    )

    actual = await llm_client.text_completion(prompt="These are a few of my favorite", max_tokens=10,)

    assert actual == ["completion text"]
    mock_aioresponse.assert_called_once_with(complete_url, method='POST',
                                             headers={AUTH_HEADER: llm_client._api_key,
                                                      ACCEPT_HEADER: ACCEPT_VALUE,
                                                      VERSION_HEADER: anthropic_version},
                                             json={PROMPT_KEY: 'These are a few of my favorite',
                                                   MAX_TOKENS_KEY: 10, "temperature": 1,
                                                   MODEL_KEY: llm_client._default_model},
                                             raise_for_status=True)


@pytest.mark.asyncio
async def test_text_completion__with_version_header(mock_aioresponse, config, complete_url):
    mock_aioresponse.post(
        complete_url,
        payload={COMPLETIONS_KEY: "completion text"}
    )
    config.headers[VERSION_HEADER] = "1.0.0"
    llm_client = AnthropicClient(config)

    actual = await llm_client.text_completion(prompt="These are a few of my favorite", max_tokens=10)

    assert actual == ["completion text"]
    mock_aioresponse.assert_called_once_with(complete_url, method='POST',
                                             headers={AUTH_HEADER: llm_client._api_key,
                                                      ACCEPT_HEADER: ACCEPT_VALUE,
                                                      VERSION_HEADER: "1.0.0"},
                                             json={PROMPT_KEY: 'These are a few of my favorite',
                                                   MAX_TOKENS_KEY: 10, "temperature": 1,
                                                   MODEL_KEY: llm_client._default_model},
                                             raise_for_status=True)


@pytest.mark.asyncio
async def test_text_completion__without_max_tokens_raise_value_error(mock_aioresponse, llm_client):
    with pytest.raises(ValueError):
        await llm_client.text_completion(prompt="These are a few of my favorite")


@pytest.mark.asyncio
async def test_text_completion__override_model(mock_aioresponse, llm_client, complete_url, anthropic_version):
    new_model_name = "claude-instant"
    mock_aioresponse.post(
        complete_url,
        payload={COMPLETIONS_KEY: "completion text"}
    )

    actual = await llm_client.text_completion(prompt="These are a few of my favorite", model=new_model_name,
                                              max_tokens=10)

    assert actual == ["completion text"]
    mock_aioresponse.assert_called_once_with(complete_url, method='POST',
                                             headers={AUTH_HEADER: llm_client._api_key,
                                                      ACCEPT_HEADER: ACCEPT_VALUE,
                                                      VERSION_HEADER: anthropic_version},
                                             json={PROMPT_KEY: 'These are a few of my favorite',
                                                   MAX_TOKENS_KEY: 10, "temperature": 1,
                                                   MODEL_KEY: new_model_name},
                                             raise_for_status=True)


@pytest.mark.asyncio
async def test_text_completion__with_kwargs(mock_aioresponse, llm_client, complete_url, anthropic_version):
    mock_aioresponse.post(
        complete_url,
        payload={COMPLETIONS_KEY: "completion text"}
    )

    actual = await llm_client.text_completion(prompt="These are a few of my favorite", max_tokens=10, temperature=0.5,top_p=0.5)

    assert actual == ["completion text"]
    mock_aioresponse.assert_called_once_with(complete_url, method='POST',
                                             headers={AUTH_HEADER: llm_client._api_key,
                                                      ACCEPT_HEADER: ACCEPT_VALUE,
                                                      VERSION_HEADER: anthropic_version},
                                             json={PROMPT_KEY: 'These are a few of my favorite',
                                                   MAX_TOKENS_KEY: 10,
                                                   MODEL_KEY: llm_client._default_model,
                                                   "temperature": 0.5, "top_p" : 0.5},
                                             raise_for_status=True)


@pytest.mark.asyncio
async def test_get_tokens_count__sanity(llm_client, number_of_tokens, mock_anthropic):
    actual = await llm_client.get_tokens_count(text="These are a few of my favorite things!")

    assert actual == 10
    mock_anthropic.return_value.count_tokens.assert_awaited_once_with("These are a few of my favorite things!")



import logging

def setup_logger(logger_name: str = "logger", log_file: str = None):
    logger = logging.getLogger(logger_name)
    logger.setLevel(logging.INFO)

    # StreamHandler for logging to console
    stream_handler = logging.StreamHandler()
    stream_handler.setFormatter(logging.Formatter('%(asctime)s - %(levelname)s - %(message)s'))
    logger.addHandler(stream_handler)

    # Optionally, log to a file
    if log_file:
        file_handler = logging.FileHandler(log_file)
        file_handler.setFormatter(logging.Formatter('%(asctime)s - %(levelname)s - %(message)s'))
        logger.addHandler(file_handler)

    return logger, stream_handler



from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from enum import Enum
from typing import Any, Optional

from dataclasses_json import dataclass_json, config

# try:
#     from aiohttp import ClientSession
# except ImportError:
#     ClientSession = Any

from sfn_llm_client import BaseLLMClient
from sfn_llm_client.utils.consts import MODEL_KEY


class Role(Enum):
    SYSTEM = "system"
    USER = "user"
    ASSISTANT = "assistant"


@dataclass_json
@dataclass
class ChatMessage:
    role: Role = field(metadata=config(encoder=lambda role: role.value, decoder=Role))
    content: str
    name: Optional[str] = field(default=None, metadata=config(exclude=lambda name: name is None))
    example: bool = field(default=False, metadata=config(exclude=lambda _: True))


@dataclass
class LLMAPIClientConfig:
    api_key: str
    # session: ClientSession
    base_url: Optional[str] = None
    default_model: Optional[str] = None
    headers: dict[str, Any] = field(default_factory=dict)


class BaseLLMAPIClient(BaseLLMClient, ABC):
    def __init__(self, config: LLMAPIClientConfig):
        self._api_key: str = config.api_key
        # self._session: ClientSession = config.session
        self._base_url: str = config.base_url
        self._default_model: str = config.default_model
        self._headers: dict[str, str] = config.headers

    @abstractmethod
    async def text_completion(self, prompt: str, model: Optional[str] = None, max_tokens: Optional[int] = None,
                              temperature: Optional[float] = None, top_p: Optional[float] = None, **kwargs) -> list[str]:
        raise NotImplementedError()

    async def chat_completion(self, messages: list[ChatMessage], temperature: float = 0,
                              max_tokens: int = 16, model: Optional[str] = None, **kwargs) -> list[str]:
        raise NotImplementedError()

    async def embedding(self, text: str, model: Optional[str] = None, **kwargs) -> list[float]:
        raise NotImplementedError()

    async def get_chat_tokens_count(self, messages: list[ChatMessage], **kwargs) -> int:
        raise NotImplementedError()

    def _set_model_in_kwargs(self, kwargs, model: Optional[str]) -> None:
        if model is not None:
            kwargs[MODEL_KEY] = model
        kwargs.setdefault(MODEL_KEY, self._default_model)




__version__ = "0.2.0a1"  


# load utils
try:
    from .utils.base_llm_client import BaseLLMClient
    from .utils.consts import BaseLLMClient
    from .utils.base_llm_client import BaseLLMClient
    from .utils.base_llm_client import BaseLLMClient

except ImportError:
    pass

# load api clients
try:
    from .llm_api_client.base_llm_api_client import BaseLLMAPIClient, LLMAPIClientConfig, ChatMessage, Role
    from .llm_api_client.llm_api_client_factory import LLMAPIClientFactory, LLMAPIClientType
    # load base-api clients
    try:
        from .llm_api_client.ai21_client import AI21Client
        from .llm_api_client.aleph_alpha_client import AlephAlphaClient
        from .llm_api_client.google_client import GoogleClient, MessagePrompt
    except ImportError:
        pass
    # load apis with different dependencies
    try:
        from .llm_api_client.openai_client import OpenAIClient
    except ImportError:
        pass
    try:
        from .llm_api_client.huggingface_client import HuggingFaceClient
    except ImportError:
        pass
    try:
        from .llm_api_client.anthropic_client import AnthropicClient
    except ImportError:
        pass
    try:
        from .llm_api_client.cortex_client import CortexClient
    except ImportError:
        pass
    try:
        from .llm_cost_calculation.openai_cost_calculation import openai_cost_calculation
    except ImportError:
        pass
    try:
        from .llm_cost_calculation.snowflake_cortex_cost_calculation import snowflake_cortex_cost_calculation
    except ImportError:
        pass
    try:
        from .llm_cost_calculation.anthropic_cost_calculation import anthropic_cost_calculation
    except ImportError:
        pass
except ImportError:
    pass
# load local clients
try:
    from .llm_client.local_client import LocalClient, LocalClientConfig
except ImportError:
    pass
# load sync support
try:
    from .sync.sync_llm_api_client_factory import init_sync_llm_api_client
except ImportError:
    pass



try:
    from .llm_api_client.core.llm import load_model
    from .llm_api_client.core.model_schema import  LLMConfig
    from .llm_cost_calculation.cost_tracker import CostCallbackHandler
except ImportError:
    pass


from typing import Any, Optional

try:
    from google.ai.generativelanguage_v1beta2 import MessagePrompt
except ImportError:
    MessagePrompt = Any  # This only needed for runtime chat_completion and chat tokens count

from sfn_llm_client.llm_api_client.base_llm_api_client import BaseLLMAPIClient, LLMAPIClientConfig
from sfn_llm_client.utils.consts import PROMPT_KEY

COMPLETE_PATH = "generateText"
CHAT_PATH = "generateMessage"
EMBEDDING_PATH = "embedText"
TOKENIZE_PATH = "countMessageTokens"
BASE_URL = "https://generativelanguage.googleapis.com/v1beta2/models/"
COMPLETIONS_KEY = "candidates"
TEXT_KEY = "text"
MAX_TOKENS_KEY = "maxOutputTokens"
COMPLETIONS_OUTPUT_KEY = "output"
TOKENS_KEY = "tokenCount"
MESSAGES_KEY = "messages"
MESSAGE_CONTENT_KEY = "content"
EMBEDDING_KEY = "embedding"
EMBEDDING_VALUE_KEY = "value"
AUTH_PARAM = "key"


class GoogleClient(BaseLLMAPIClient):
    def __init__(self, config: LLMAPIClientConfig):
        super().__init__(config)
        if self._base_url is None:
            self._base_url = BASE_URL
        self._params = {AUTH_PARAM: self._api_key}

    async def text_completion(self, prompt: str, model: Optional[str] = None, max_tokens: Optional[int] = 64,
                              temperature: Optional[float] = None,top_p: Optional[float] = None, **kwargs) -> list[str]:
        model = model or self._default_model
        kwargs[PROMPT_KEY] = {TEXT_KEY: prompt}
        kwargs[MAX_TOKENS_KEY] = kwargs.pop(MAX_TOKENS_KEY, max_tokens)
        if top_p:
            kwargs["topP"] = top_p
        kwargs["temperature"] = kwargs.pop("temperature", temperature)
        response = await self._session.post(self._base_url + model + ":" + COMPLETE_PATH,
                                            params=self._params,
                                            json=kwargs,
                                            headers=self._headers,
                                            raise_for_status=True)
        response_json = await response.json()
        completions = response_json[COMPLETIONS_KEY]
        return [completion[COMPLETIONS_OUTPUT_KEY] for completion in completions]

    async def embedding(self, text: str, model: Optional[str] = None, **kwargs) -> list[float]:
        model = model or self._default_model
        response = await self._session.post(self._base_url + model + ":" + EMBEDDING_PATH,
                                            params=self._params,
                                            json={TEXT_KEY: text},
                                            headers=self._headers,
                                            raise_for_status=True)
        response_json = await response.json()
        return response_json[EMBEDDING_KEY][EMBEDDING_VALUE_KEY]

    async def get_tokens_count(self, text: str, model: Optional[str] = None,
                               messages: Optional[MessagePrompt] = None, **kwargs) -> int:
        """
        Retrieves the count of tokens in the given text using the specified model or the default_model.

        :param text: (str) The input text to tokenize and count the tokens. If the keyword argument `${MESSAGES_KEY}` \
                       is provided, this parameter will be ignored.
        :param model: (Optional[str], optional) The name of the model to use for tokenization. If not provided,
                       the default model will be used. Defaults to `None`.
        :param messages: (MessagePrompt | None, optional) The messages to tokenize and count the tokens. If provided,
                            the `text` parameter will be ignored.
        :param kwargs: Ignored.
        :return: (int) The count of tokens in the given text.
        """

        model = model or self._default_model
        if not messages:
            messages = {MESSAGES_KEY: [{MESSAGE_CONTENT_KEY: text}]}
        response = await self._session.post(self._base_url + model + ":" + TOKENIZE_PATH,
                                            params=self._params,
                                            json={PROMPT_KEY: messages},
                                            headers=self._headers,
                                            raise_for_status=True)
        response_json = await response.json()
        return response_json[TOKENS_KEY]



from typing import Optional
from langchain_openai import ChatOpenAI
from sfn_llm_client.llm_api_client.base_llm_api_client import BaseLLMAPIClient, LLMAPIClientConfig, ChatMessage
from sfn_llm_client.utils.consts import PROMPT_KEY
from sfn_llm_client.llm_cost_calculation.openai_cost_calculation import openai_cost_calculation
from sfn_llm_client.utils.logging import setup_logger
from sfn_llm_client.utils.retry_with import retry_with

class OpenAILangchainClient(BaseLLMAPIClient):
    def __init__(self, config: LLMAPIClientConfig):
        super().__init__(config)
        self.logger, _ = setup_logger(logger_name="OpenAILangchainClient")
        self._client = None # Stores the single ChatOpenAI client instance
        self._client_params = None # Stores the parameters of the current client
    
    def get_or_create_client(self, api_key: str, model: str, temperature: float, max_tokens: int):
        """
        Gets the existing ChatOpenAI client or creates a new one if it doesn't exist
        or if the requested parameters are different from the current client's parameters.

        Args:
            api_key (str): The OpenAI API key.
            model (str): The OpenAI model name (e.g., "gpt-3.5-turbo").
            temperature (float): The sampling temperature.
            max_tokens (int): The maximum number of tokens to generate.

        Returns:
            ChatOpenAI: An instance of the ChatOpenAI client.
        """
        requested_params = (api_key, model, temperature, max_tokens)

        # Check if a client already exists and if its parameters match the requested ones.
        if self._client and self._client_params == requested_params:
            print(f"Reusing existing client with parameters: {requested_params}")
            return self._client
        else:
            # If no client exists, or if parameters are different, create a new one.
            print(f"Creating new client for parameters: {requested_params}")
            try:
                # temperature not supported in following models
                if model in ("o1-mini", "o1-pro", "o3", "o4-mini", "o3-mini"):
                    client = ChatOpenAI(
                    api_key=api_key,
                    model=model,
                    max_tokens=max_tokens
                    )
                else:
                    client = ChatOpenAI(
                        api_key=api_key,
                        model=model,
                        temperature=temperature,
                        max_tokens=max_tokens
                    )
                self._client = client
                self._client_params = requested_params
                return self._client
            except Exception as e:
                print(f"Error creating ChatOpenAI client: {e}")
                raise
    # TODO: try if get_or_create_client can be used instead of get_langchain_llm
    def get_langchain_llm(self, api_key=None, model=None, temperature=None, max_tokens=None):
        """
        Returns the underlying ChatOpenAI client, creating it if necessary.
        """
        api_key = api_key or self._api_key
        model = model or self._default_model
        temperature = temperature if temperature is not None else 0
        max_tokens = max_tokens if max_tokens is not None else 16
        return self.get_or_create_client(api_key=api_key, model=model, temperature=temperature, max_tokens=max_tokens)

    @retry_with(retries=3, retry_delay=3.0, backoff=True)
    def chat_completion(self, messages: list[ChatMessage], temperature: float = 0,
                        max_tokens: int = 16, top_p: float = 1, model: Optional[str] = None, 
                        retries: int = 3, retry_delay: float = 3.0, **kwargs) -> list[str]:
        """
        This method performs chat completion with OpenAI, and includes basic retry logic for handling
        exceptions or empty responses.

        :param retries: Number of retries in case of failure.
        :param retry_delay: Delay in seconds between retries.
        """
        self._set_model_in_kwargs(kwargs, model)
        messages = [
            message if isinstance(message, dict) else message.to_dict() 
            for message in messages
        ]
        self.get_or_create_client(api_key=self._api_key, model=model, temperature=temperature, max_tokens=max_tokens)
        
        completions = self._client.invoke(messages)
        print("completions:" , completions)

        # Check if response is empty
        if not completions or not completions.content:
            raise ValueError("Received empty response from the openai llm")

        token_usage = completions.response_metadata["token_usage"]
        prompt_tokens = token_usage['prompt_tokens']
        completion_tokens = token_usage['completion_tokens']
        
        token_cost_summary = openai_cost_calculation(
            prompt_tokens,
            completion_tokens,
            model,
        )
        return completions, token_cost_summary
    
    async def text_completion(self, *args, **kwargs):
        raise NotImplementedError("text_completion is not supported in OpenAILangchainClient.")

    async def embedding(self, *args, **kwargs):
        raise NotImplementedError("embedding is not supported in OpenAILangchainClient.")

    async def get_chat_tokens_count(self, *args, **kwargs):
        raise NotImplementedError("get_chat_tokens_count is not supported in OpenAILangchainClient.")


from unittest.mock import patch, AsyncMock

import pytest

from sfn_llm_client.llm_api_client.anthropic_client import BASE_URL, COMPLETE_PATH, VERSION_HEADER, AnthropicClient
from sfn_llm_client.llm_api_client.base_llm_api_client import LLMAPIClientConfig


@pytest.fixture
def model_name():
    return "claude-v1"


@pytest.fixture
def config(client_session, model_name):
    return LLMAPIClientConfig("top-secret-api-key", client_session, default_model=model_name)


@pytest.fixture
def llm_client(config):
    return AnthropicClient(config)


@pytest.fixture
def complete_url():
    return BASE_URL + COMPLETE_PATH


@pytest.fixture
def number_of_tokens():
    return 10


@pytest.fixture
def anthropic_version():
    return "2023-06-01"


@pytest.fixture(autouse=True)
def mock_anthropic(number_of_tokens, anthropic_version):
    with patch("sfn_llm_client.llm_api_client.anthropic_client.AsyncAnthropic") as mock_anthropic:
        mock_anthropic.return_value.count_tokens = AsyncMock(return_value=number_of_tokens)
        mock_anthropic.return_value.default_headers = {VERSION_HEADER: anthropic_version}
        yield mock_anthropic



from typing import Optional

import pytest

from sfn_llm_client.llm_api_client.google_client import GoogleClient, BASE_URL, AUTH_PARAM
from sfn_llm_client.llm_api_client.base_llm_api_client import LLMAPIClientConfig, BaseLLMAPIClient


@pytest.fixture
def model_name():
    return "text-bison-001"


@pytest.fixture
def config(client_session, model_name):
    return LLMAPIClientConfig("top-secret-api-key", client_session, default_model=model_name)


@pytest.fixture
def llm_client(config):
    return GoogleClient(config)


@pytest.fixture
def params(llm_client):
    return "?" + AUTH_PARAM + "=" + llm_client._api_key


def build_url(llm_client: BaseLLMAPIClient, path: str, model: Optional[str] = None) -> str:
    model = model or llm_client._default_model
    return BASE_URL + model + ":" + path



import json
# import re
from snowflake.cortex import Complete
from typing import Optional
from sfn_llm_client.llm_api_client.base_llm_api_client import (
    BaseLLMAPIClient,
    ChatMessage
)
from sfn_llm_client.utils.logging import setup_logger
from snowflake.snowpark import Session
from sfn_llm_client.llm_cost_calculation.snowflake_cortex_cost_calculation import snowflake_cortex_cost_calculation
from sfn_llm_client.utils.retry_with import retry_with
class CortexClient(BaseLLMAPIClient):
    def __init__(self):
        self.logger, _ = setup_logger(logger_name="SnowflakeCortex")

    @retry_with(retries=3, retry_delay=3.0, backoff=True)
    def chat_completion(
        self,
        messages: list[ChatMessage],
        temperature: float = 0,
        max_tokens: int = 16,
        top_p: float = 1,
        model: Optional[str] = "snowflake-arctic",
        retries: int = 3,
        retry_delay: float = 3.0,
        session: Optional[Session] = None,
        **kwargs,
    ) -> list[str]:
        self.logger.info('Started calling Cortex Complete API...')

        completions = Complete(
            model,
            prompt=messages,
            options={"max_tokens": max_tokens, "temperature": temperature, "guardrails": False},
            session=session,
        )

        self.logger.info(f"Received cortex {model}, Completions response...{completions}")
        
        # response_content = response['choices'][0]['messages']
        # pattern = re.compile(r'\{.*"text_response".*"mapping".*\}', re.DOTALL)
        # match = pattern.search(response_content)
        # if match:
        #     extracted_json = match.group(0)  # Extract the dictionary part
        # else:
        #     return {"text_response":"Null","mapping":{}}
        # try:
        #     response_content = json.loads(extracted_json)
        # except json.JSONDecodeError:
        #     self.error("Error: Failed to decode JSON")

        # Calculate token consumption

        token_cost_summary = snowflake_cortex_cost_calculation(
            response=completions,
            model=model
        )
        self.logger.info(f"After consumed token's cost calculation received token_cost_summary...{token_cost_summary}")

        return completions, token_cost_summary
    
    
    async def text_completion(self, *args, **kwargs):
        raise NotImplementedError("text_completion is not supported in CortexClient.")

    async def embedding(self, *args, **kwargs):
        raise NotImplementedError("embedding is not supported in CortexClient.")

    async def get_chat_tokens_count(self, *args, **kwargs):
        raise NotImplementedError("get_chat_tokens_count is not supported in CortexClient.")


from unittest.mock import call

import pytest

from sfn_llm_client.llm_client.local_client import LocalClient, LocalClientConfig


@pytest.mark.asyncio
async def test_text_completion__sanity(local_client, mock_model, mock_tokenizer, tensors_type, device):
    mock_tokenizer.encode.return_value.to.return_value = [1, 2, 3]
    mock_model.generate.return_value = [1]
    mock_tokenizer.decode.return_value = "first completion"

    actual = await local_client.text_completion(prompt="These are a few of my favorite")

    assert actual == ["first completion"]
    mock_tokenizer.encode.assert_called_once_with("These are a few of my favorite", return_tensors=tensors_type)
    mock_tokenizer.encode.return_value.to.assert_called_once_with(device)
    mock_model.generate.assert_called_once_with([1, 2, 3])
    mock_tokenizer.decode.assert_called_once_with(1)


@pytest.mark.asyncio
async def test_text_completion__return_multiple_completions(local_client, mock_model, mock_tokenizer, tensors_type,
                                                            device):
    mock_tokenizer.encode.return_value.to.return_value = [1, 2, 3]
    mock_model.generate.return_value = [2, 3, 4]
    mock_tokenizer.decode.side_effect = ["first completion", "second completion", "third completion"]

    actual = await local_client.text_completion(prompt="These are a few of my favorite")

    assert actual == ["first completion", "second completion", "third completion"]
    mock_tokenizer.encode.assert_called_once_with("These are a few of my favorite", return_tensors=tensors_type)
    mock_tokenizer.encode.return_value.to.assert_called_once_with(device)
    mock_model.generate.assert_called_once_with([1, 2, 3])
    assert mock_tokenizer.decode.call_args_list == [call(2), call(3), call(4)]


@pytest.mark.asyncio
async def test_text_completion__with_kwargs(local_client, mock_model, mock_tokenizer, tensors_type, device):
    mock_tokenizer.encode.return_value.to.return_value = [1, 2, 3]
    mock_model.generate.return_value = [1]
    mock_tokenizer.decode.return_value = "first completion"

    actual = await local_client.text_completion(prompt="These are a few of my favorite", max_length=100)

    assert actual == ["first completion"]
    mock_tokenizer.encode.assert_called_once_with("These are a few of my favorite", return_tensors=tensors_type)
    mock_tokenizer.encode.return_value.to.assert_called_once_with(device)
    mock_model.generate.assert_called_once_with([1, 2, 3], max_length=100)
    mock_tokenizer.decode.assert_called_once_with(1)


@pytest.mark.asyncio
async def test_text_completion__with_encode_kwargs(mock_model, mock_tokenizer, tensors_type, device):
    mock_tokenizer.encode.return_value.to.return_value = [1, 2, 3]
    mock_model.generate.return_value = [1]
    mock_tokenizer.decode.return_value = "first completion"
    encode_kwargs = {"add_special_tokens": False}
    local_client = LocalClient(LocalClientConfig(mock_model, mock_tokenizer, tensors_type, device, encode_kwargs))

    actual = await local_client.text_completion(prompt="These are a few of my favorite")

    assert actual == ["first completion"]
    mock_tokenizer.encode.assert_called_once_with("These are a few of my favorite", return_tensors=tensors_type,
                                                  add_special_tokens=False)
    mock_tokenizer.encode.return_value.to.assert_called_once_with(device)
    mock_model.generate.assert_called_once_with([1, 2, 3])
    mock_tokenizer.decode.assert_called_once_with(1)


@pytest.mark.asyncio
async def test_get_tokens_count__sanity(local_client, mock_tokenizer, tensors_type, device):
    mock_tokenizer.encode.return_value.to.return_value = [1, 2, 3]

    actual = await local_client.get_tokens_count(text="This is a test")

    assert actual == 3
    mock_tokenizer.encode.assert_called_once_with("This is a test", return_tensors=tensors_type)
    mock_tokenizer.encode.return_value.to.assert_called_once_with(device)


@pytest.mark.asyncio
async def test_get_tokens_count__with_kwargs(mock_model, mock_tokenizer, tensors_type, device):
    mock_tokenizer.encode.return_value.to.return_value = [1, 2, 3]
    encode_kwargs = {"add_special_tokens": False}
    local_client = LocalClient(LocalClientConfig(mock_model, mock_tokenizer, tensors_type, device, encode_kwargs))

    actual = await local_client.get_tokens_count(text="This is a test")

    assert actual == 3
    mock_tokenizer.encode.assert_called_once_with("This is a test", return_tensors=tensors_type,
                                                  add_special_tokens=False)
    mock_tokenizer.encode.return_value.to.assert_called_once_with(device)



from abc import ABC, abstractmethod


class BaseLLMClient(ABC):
    @abstractmethod
    async def text_completion(self, prompt: str, **kwargs) -> list[str]:
        raise NotImplementedError()

    async def get_tokens_count(self, text: str, **kwargs) -> int:
        raise NotImplementedError()



import inspect
from functools import wraps
from typing import Callable, Any

import async_to_sync
from aiohttp import ClientSession

from sfn_llm_client import LLMAPIClientConfig
from sfn_llm_client.llm_api_client.llm_api_client_factory import LLMAPIClientType, get_llm_api_client_class


def _create_new_session(f: Callable[..., Any]) -> Callable[..., Any]:
    @wraps(f)
    async def func(self, *args, **kwargs):
        self._session = ClientSession()
        try:
            response = await f(self, *args, **kwargs)
        except Exception as e:
            await self._session.close()
            raise e
        await self._session.close()
        return response

    return func


def _decorate_all_methods_in_class(decorators):
    def apply_decorator(cls: Any) -> Any:
        for k, f in cls.__dict__.items():
            if inspect.isfunction(f) and not k.startswith("_"):
                for decorator in decorators:
                    setattr(cls, k, decorator(cls.__dict__[k]))
        return cls

    return apply_decorator


def init_sync_llm_api_client(llm_api_client_type: LLMAPIClientType, **config_kwargs) -> "Sync BaseLLMAPIClient":
    llm_api_client_class = get_llm_api_client_class(llm_api_client_type)
    return async_to_sync.methods(_decorate_all_methods_in_class([_create_new_session])(llm_api_client_class)(
        LLMAPIClientConfig(session=None, **config_kwargs)))



from sfn_llm_client.utils.consts import ANTHROPIC_DEFAULT_MODEL, ANTHROPIC_MODEL_TOKENS_COST
from sfn_llm_client.utils.logging import setup_logger

# Initialize logger
logger, _ = setup_logger(logger_name="AnthropicCostCalculation")

def anthropic_cost_calculation(
    total_input_tokens: int,
    total_output_tokens: int,
    model: str = ANTHROPIC_DEFAULT_MODEL
) -> dict:
    """
    This method calculates usage cost of Anthropic model, based on the number of tokens used.

    Args:
        total_input_tokens: The total number of tokens in the input.
        total_output_tokens: The total number of tokens in the model's response.
        model: The name of the Anthropic model to use or Defaults to ANTHROPIC_DEFAULT_MODEL.

    Returns:
        A dictionary containing the following keys:
            - prompt_tokens: The total number of input tokens.
            - completions_tokens: The total number of output tokens.
            - total_tokens: The total number of tokens.
            - total_cost_usd: The total cost in USD.
    """
    logger.info(f'started Anthropic cost calculation with model: {model}')

    try:
        model_pricing = ANTHROPIC_MODEL_TOKENS_COST[model]
    except KeyError:
        raise ValueError(f"Invalid model specified: {model}")

    # Calculate costs
    input_cost = total_input_tokens * model_pricing["input"] / 1000
    output_cost = total_output_tokens * model_pricing["output"] / 1000
    total_tokens = total_input_tokens + total_output_tokens
    total_cost_usd = input_cost + output_cost

    token_cost_summary = {
        "prompt_tokens": total_input_tokens,
        "completion_tokens": total_output_tokens,
        "total_tokens": total_tokens,
        "total_cost_usd": round(total_cost_usd, 4),
    }
    logger.info(f'Anthropic cost calculation done! total cost is: {token_cost_summary}')
    return token_cost_summary


import pytest
import pytest_asyncio
from aiohttp import ClientSession
from aioresponses import aioresponses


@pytest.fixture
def mock_aioresponse():
    with aioresponses() as m:
        yield m


@pytest_asyncio.fixture
async def client_session():
    session = ClientSession()
    yield session
    await session.close()



from snowflake.snowpark import Session
from langchain_community.chat_models.snowflake import ChatSnowflakeCortex
from langchain_core.messages import SystemMessage, HumanMessage, AIMessage
from pydantic import model_validator
from typing import Dict
from sfn_llm_client.llm_api_client.base_llm_api_client import ChatMessage
from typing import Optional
from sfn_llm_client.utils.consts import PROMPT_KEY
from sfn_llm_client.llm_cost_calculation.snowflake_cortex_cost_calculation import snowflake_cortex_cost_calculation_langchain
from sfn_llm_client.llm_cost_calculation.openai_cost_calculation import openai_cost_calculation
from sfn_llm_client.utils.logging import setup_logger
from sfn_llm_client.utils.retry_with import retry_with

# Custom class that inherits from ChatSnowflakeCortex to override methods.
class CustomChatSnowflakeCortex(ChatSnowflakeCortex):
    """
    A custom ChatSnowflakeCortex class to demonstrate overriding the
    environment validation logic.
    """
    @model_validator(mode="before")
    def validate_environment(cls, values: Dict) -> Dict:
        """
        Overrides the default pydantic environment validation.
        The original validator ensures a snowpark_session is present,
        which is a critical check we will maintain.
        """
        print("Executing custom override of validate_environment.")
        print("DEBUG: validate_environment() received values:", values)
        session = values.get("session")
        if session and isinstance(session, Session):
            print(f"Type of session: {type(session)}")
            print("DEBUG: Existing Snowpark session found. Skipping new session creation.")
            return values

        # If no session is provided, let the original parent validator handle the creation using environment variables.
        print("DEBUG: No session found. Using parent validation to create one.")
        return super().validate_environment(values)

class CortexLangchainClient():
    
    def __init__(self):
       
        self.logger, _ = setup_logger(logger_name="CortexLangchainClient")

    @retry_with(retries=3, retry_delay=3.0, backoff=True)
    def chat_completion(self,
        messages: list[ChatMessage],
        temperature: float = 0,
        max_tokens: int = 16,
        top_p: float = 1,
        model: Optional[str] = "snowflake-arctic",
        retries: int = 3,
        retry_delay: float = 3.0,
        session: Optional[Session] = None,
        **kwargs,
    ) -> list[str]:
        """
        Generates a response from Snowflake Cortex using the custom chat model.
        """
        client = CustomChatSnowflakeCortex(
            model=model,
            cortex_function="try_complete",
            temperature=temperature,
            max_tokens=max_tokens,
            top_p=top_p,
            session=session,
            snowflake_username="oauth_user",  # Dummy username to pass validation
        )
        try:
            cleaned_messages = self.convert_messages(messages)
            completions = client.invoke(cleaned_messages)
            print("completions: ", completions)

            # Check if response is empty
            if not completions or not completions.content:
                raise ValueError("Received empty response from the openai llm")

            print("completions:" , completions)

            token_cost_summary = snowflake_cortex_cost_calculation_langchain(
            response_metadata=completions.response_metadata,
            model=model
            )
            self.logger.info(f"After consumed token's cost calculation received token_cost_summary...{token_cost_summary}")

            return completions, token_cost_summary
        except Exception as e:
            print(f"Error during Snowflake Cortex generation: {e}")
            return f"Error: Could not process request with Snowflake Cortex. {e}", None
        
    def convert_messages(self, raw_messages: list[dict]) -> list:
        """
        Convert dict-based messages to LangChain message objects with cleaning.
        Supports 'system', 'user', and 'assistant' roles.
        """
        role_to_class = {
            "system": SystemMessage,
            "user": HumanMessage,
            "assistant": AIMessage
        }

        converted = []
        for msg in raw_messages:
            role = msg.get("role", "").lower()
            content = self.clean_prompt(msg.get("content", ""))
            msg_class = role_to_class.get(role)
            if msg_class:
                converted.append(msg_class(content=content))
            else:
                print(f"Skipping unsupported role: {role}")
        return converted
    
    def clean_prompt(self, raw_prompt: str) -> str:
        """
        Clean prompt to avoid SQL syntax errors and ensure JSON compatibility.
        This approach focuses on escaping characters that would break SQL without
        interfering with JSON string integrity.
        """
        if not isinstance(raw_prompt, str):
            return ""
        cleaned_prompt = raw_prompt.rstrip("\n").splitlines()
        cleaned_prompt = " ".join(cleaned_prompt).replace("'", " ").replace('"', " ")
        return cleaned_prompt

    def get_langchain_llm(self, model="snowflake-arctic", temperature=0, max_tokens=16, top_p=1, session=None):
        """
        Returns a CustomChatSnowflakeCortex instance for agentic workflows.
        """
        return CustomChatSnowflakeCortex(
            model=model,
            cortex_function="try_complete",
            temperature=temperature,
            max_tokens=max_tokens,
            top_p=top_p,
            session=session,
            snowflake_username="oauth_user",
        )


MODEL_KEY = "model"
PROMPT_KEY = "prompt"
OPENAI_DEFAULT_MODEL = "gpt-3.5-turbo-0125"
ANTHROPIC_DEFAULT_MODEL = "claude-3-5-sonnet-20241022"

CORTEX_MODEL_TOKENS_COST = {
    "gemma-7b": {"TOKENS_COST": 0.12},
    "jamba-instruct": {"TOKENS_COST": 0.83},
    "jamba-1.5-large": {"TOKENS_COST": 1.40},
    "jamba-1.5-mini": {"TOKENS_COST": 0.10},
    "llama3.1-405b": {"TOKENS_COST": 3.00},
    "llama3.1-70b": {"TOKENS_COST": 1.21},
    "llama3.1-8b": {"TOKENS_COST": 0.19},
    "llama3.2-1b": {"TOKENS_COST": 0.04},
    "llama3.2-3b": {"TOKENS_COST": 0.06},
    "mistral-large2": {"TOKENS_COST": 1.95},
    "mistral-7b": {"TOKENS_COST": 0.12},
    "mixtral-8x7b": {"TOKENS_COST": 0.22},
    "reka-core": {"TOKENS_COST": 5.50},
    "reka-flash": {"TOKENS_COST": 0.45},
    "snowflake-arctic": {"TOKENS_COST": 0.84},
    "claude-3-5-sonnet": {"TOKENS_COST": 2.55}
}

# pricing for 1k tokens
OPENAI_MODEL_TOKENS_COST = {
        "gpt-3.5-turbo-0125": {
            "prompt": 0.0005,
            "completion": 0.0015,
        },
        "gpt-3.5-turbo-16k": {
            "prompt": 0.003,
            "completion": 0.004,
        },
        "gpt-3.5-turbo": {
            "prompt": 0.003,
            "completion": 0.006,
        },
        "gpt-4-8k": {
            "prompt": 0.03,
            "completion": 0.06,
        },
        "gpt-4-32k": {
            "prompt": 0.06,
            "completion": 0.12,
        },
        "text-embedding-ada-002-v2": {
            "prompt": 0.0001,
            "completion": 0.0001,
        },
        "gpt-4.1": {
            "prompt": 0.002,
            "completion": 0.008
        },
        "gpt-4.1-mini": {
            "prompt": 0.0004,
            "completion": 0.0016
        },
        "gpt-4.1-nano": {
            "prompt": 0.0001,
            "completion": 0.0004
        },
        "gpt-4.5-preview": {
            "prompt": 0.075,
            "completion": 0.15
        },
        "gpt-4o": {
            "prompt": 0.0025,
            "completion": 0.01
        },
        "gpt-4o-audio-preview": {
            "prompt": 0.0025,
            "completion": 0.01
        },
        "gpt-4o-realtime-preview": {
            "prompt": 0.005,
            "completion": 0.02
        },
        "gpt-4o-mini": {
            "prompt": 0.00015,
            "completion": 0.0006
        },
        "gpt-4o-mini-audio-preview": {
            "prompt": 0.00015,
            "completion": 0.0006
        },
        "gpt-4o-mini-realtime-preview": {
            "prompt": 0.0006,
            "completion": 0.0024
        },
        "o1": {
            "prompt": 0.015,
            "completion": 0.06
        },
        "o1-pro": {
            "prompt": 0.15,
            "completion": 0.6
        },
        "o3-pro": {
            "prompt": 0.02,
            "completion": 0.08
        },
        "o3": {
            "prompt": 0.002,
            "completion": 0.008
        },
        "o4-mini": {
            "prompt": 0.0011,
            "completion": 0.0044
        },
        "o3-mini": {
            "prompt": 0.0011,
            "completion": 0.0044
        },
        "o1-mini": {
            "prompt": 0.0011,
            "completion": 0.0044
        },
        "codex-mini-latest": {
            "prompt": 0.0015,
            "completion": 0.006
        },
        "gpt-4o-mini-search-preview": {
            "prompt": 0.00015,
            "completion": 0.0006
        },
        "gpt-4o-search-preview": {
            "prompt": 0.0025,
            "completion": 0.01
        },
        "computer-use-preview": {
            "prompt": 0.003,
            "completion": 0.012
        },
    }

# cost token per million
ANTHROPIC_MODEL_TOKENS_COST = {
    "claude-3-5-sonnet-20241022": {"input": 3.00, "output": 15.00},
    "claude-3-5-sonnet-20240620": {"input": 3.00, "output": 15.00},
    "claude-3-5-haiku-20241022": {"input": 1.00, "output": 5.00},
    "claude-3-opus-20240229": {"input": 15.00, "output": 75.00},
    "claude-3-sonnet-20240229": {"input": 3.00, "output": 15.00},
    "claude-3-haiku-20240307": {"input": 0.25, "output": 1.25}
}


from dataclasses import dataclass, field
from typing import Any

try:
    from transformers import PreTrainedModel, PreTrainedTokenizerBase
except ImportError:
    PreTrainedModel = Any
    PreTrainedTokenizerBase = Any

from sfn_llm_client import BaseLLMClient


@dataclass
class LocalClientConfig:
    model: PreTrainedModel
    tokenizer: PreTrainedTokenizerBase
    tensors_type: str
    device: str
    encode_kwargs: dict[str, Any] = field(default_factory=dict)


class LocalClient(BaseLLMClient):
    def __init__(self, llm_client_config: LocalClientConfig):
        if not llm_client_config.model.can_generate():
            raise TypeError(f"{llm_client_config.model} is not a text generation model")

        self._model: PreTrainedModel = llm_client_config.model
        self._tokenizer: PreTrainedTokenizerBase = llm_client_config.tokenizer
        self._tensors_type: str = llm_client_config.tensors_type
        self._device: str = llm_client_config.device
        self._encode_kwargs: dict[str, Any] = llm_client_config.encode_kwargs
        self._encode_kwargs["return_tensors"] = llm_client_config.tensors_type

    async def text_completion(self, prompt: str, **kwargs) -> list[str]:
        input_ids = self._encode(prompt)
        outputs = self._model.generate(input_ids, **kwargs)
        return [self._tokenizer.decode(output) for output in outputs]

    async def get_tokens_count(self, text: str, **kwargs) -> int:
        return len(self._encode(text))

    def _encode(self, prompt: str) -> list[int]:
        return self._tokenizer.encode(prompt, **self._encode_kwargs).to(self._device)



import json
import os
from pathlib import Path


def load_json_resource(path: str) -> dict:
    project_base_path = Path(__file__).parent.parent
    with open(os.path.join(f'{project_base_path}/resources/', path), "r") as f:
        return json.load(f)



from unittest.mock import MagicMock

import pytest
from transformers import PreTrainedTokenizerBase

from sfn_llm_client import LocalClientConfig, LocalClient


@pytest.fixture
def mock_model():
    return MagicMock()


@pytest.fixture
def mock_tokenizer():
    return MagicMock(PreTrainedTokenizerBase)


@pytest.fixture
def tensors_type():
    return "pt"


@pytest.fixture
def device():
    return "cpu"


@pytest.fixture
def local_client(mock_model, mock_tokenizer, tensors_type, device):
    return LocalClient(LocalClientConfig(mock_model, mock_tokenizer, tensors_type, device))



from sfn_llm_client.utils.consts import CORTEX_MODEL_TOKENS_COST
import tiktoken
import json

def calculate_tokens(text: str, model: str) -> int:
    """Calculate the number of tokens in the given text."""
    encoding = tiktoken.encoding_for_model(model)
    return len(encoding.encode(text))

def snowflake_cortex_cost_calculation(response: dict, model: str) -> tuple:
    """Calculate the cost for consumed tokens for cortex llm."""
    # In Cortex prompt and completions both tokens has same cost/credits
    # So keeping the sum as total of tokens to calculate dollar bill
    response = json.loads(response)
    # Check if response is empty
    if not response or not response['choices']:
        raise ValueError("Received empty response from the openai llm")

    # Ensure model is supported
    if model not in CORTEX_MODEL_TOKENS_COST:
        raise ValueError(f"Unsupported model: {model}")

    # Extract token usage from response
    prompt_tokens = response['usage']['prompt_tokens']
    completion_tokens = response['usage']['completion_tokens']
    guardrails_tokens = response['usage'].get('guardrails_tokens', 0)  # Handle missing guardrails_tokens
    
    # Calculate total tokens and cost
    total_tokens = prompt_tokens + completion_tokens + guardrails_tokens
    token_cost_in_credits = total_tokens * CORTEX_MODEL_TOKENS_COST[model]["TOKENS_COST"] / 1000000  # Cost in credits

    token_cost_summary = {
        "prompt_tokens": prompt_tokens,
        "completion_tokens": completion_tokens,
        "guardrails_tokens": guardrails_tokens,
        "total_tokens": total_tokens,
        "total_cost_in_credits": token_cost_in_credits
    }
    
    return token_cost_summary


def snowflake_cortex_cost_calculation_langchain(response_metadata: dict, model: str) -> tuple:
    """Calculate the cost for consumed tokens for cortex llm."""
    
    # Check if response_metadata is empty
    if not response_metadata:
        raise ValueError("Received empty response_metadata from the cortex langchain llm")

    # Ensure model is supported
    if model not in CORTEX_MODEL_TOKENS_COST:
        raise ValueError(f"Unsupported model: {model}")

    # Extract token usage from response_metadata
    prompt_tokens = response_metadata['prompt_tokens']
    completion_tokens = response_metadata['completion_tokens']
    guardrails_tokens = response_metadata.get('guardrails_tokens', 0)  # Handle missing guardrails_tokens
    
    # Calculate total tokens and cost
    total_tokens = prompt_tokens + completion_tokens + guardrails_tokens
    token_cost_in_credits = total_tokens * CORTEX_MODEL_TOKENS_COST[model]["TOKENS_COST"] / 1000000  # Cost in credits

    token_cost_summary = {
        "prompt_tokens": prompt_tokens,
        "completion_tokens": completion_tokens,
        "guardrails_tokens": guardrails_tokens,
        "total_tokens": total_tokens,
        "total_cost_in_credits": token_cost_in_credits
    }
    
    return token_cost_summary


from unittest.mock import MagicMock, patch

import pytest
from aiohttp import ClientSession

from sfn_llm_client.llm_api_client.base_llm_api_client import LLMAPIClientConfig
from sfn_llm_client.llm_api_client.openai_client import OpenAIClient


@pytest.fixture
def model_name():
    return "ada"


@pytest.fixture
def openai_mock():
    with patch("sfn_llm_client.llm_api_client.openai_client.openai") as openai_mock:
        yield openai_mock


@pytest.fixture
def config(model_name):
    return LLMAPIClientConfig("fake-api-key", MagicMock(ClientSession), default_model=model_name)


@pytest.fixture
def open_ai_client(config):
    return OpenAIClient(config)


@pytest.fixture
def tiktoken_mock():
    with patch("sfn_llm_client.llm_api_client.openai_client.tiktoken") as tiktoken_mock:
        yield tiktoken_mock



import time
from functools import wraps
from typing import Callable, Any
from sfn_llm_client.utils.logging import setup_logger

def retry_with(
    retries: int = 3,
    retry_delay: float = 3.0,
    backoff: bool = False
) -> Callable:
    """
    A Retry decorator
    Args:
        retries: Maximum number of retry attempts.
        retry_delay: Delay between retries in seconds.
        backoff: If True, applies exponential backoff to retry delay.
        logger: Logger instance for logging retries and errors.
    """
    logger, _ = setup_logger(logger_name="RetryDecorator")

    def decorator(func: Callable) -> Callable:
        @wraps(func)
        def wrapper(*args, **kwargs) -> Any:
            delay = retry_delay
            for attempt in range(retries):
                try:
                    return func(*args, **kwargs)
                except Exception as e:
                    if attempt < retries - 1:
                        logger.warning(
                            f"Attempt {attempt + 1}/{retries} failed: {str(e)}. Retrying in {delay} seconds..."
                        )
                        time.sleep(delay)
                        if backoff:
                            delay *= 2  # Exponential backoff
                    else:
                        logger.error(
                            f"Attempt {attempt + 1}/{retries} failed: {str(e)}. No more retries left."
                        )
                        raise
        return wrapper
    return decorator



import pytest

from sfn_llm_client import LLMAPIClientType, LLMAPIClientFactory
from sfn_llm_client.llm_api_client.huggingface_client import AUTH_HEADER, \
    BEARER_TOKEN, HuggingFaceClient
from tests.test_utils.load_json_resource import load_json_resource


@pytest.mark.asyncio
async def test_get_llm_api_client__with_hugging_face(config):
    del config.session
    async with LLMAPIClientFactory() as llm_api_client_factory:

        actual = llm_api_client_factory.get_llm_api_client(LLMAPIClientType.HUGGING_FACE, **config.__dict__)

    assert isinstance(actual, HuggingFaceClient)


@pytest.mark.asyncio
async def test_text_completion__sanity(mock_aioresponse, llm_client, url):
    mock_aioresponse.post(
        url,
        payload=load_json_resource("huggingface/text_completion.json")
    )

    actual = await llm_client.text_completion(prompt="who is kobe bryant")

    assert actual == ['Kobe Bryant is a retired professional basketball player who played for the Los Angeles Lakers of']
    mock_aioresponse.assert_called_once_with(url, method='POST',
                                             headers={AUTH_HEADER: BEARER_TOKEN + llm_client._api_key},
                                             json={'inputs': 'who is kobe bryant',"max_length": None, "temperature": 1.0, "top_p" : None},
                                             raise_for_status=True)


@pytest.mark.asyncio
async def test_text_completion__with_kwargs(mock_aioresponse, llm_client, url):
    mock_aioresponse.post(
        url,
        payload=load_json_resource("huggingface/text_completion.json")
    )

    actual = await llm_client.text_completion(prompt="who is kobe bryant",max_tokens = 10)

    assert actual == ['Kobe Bryant is a retired professional basketball player who played for the Los Angeles Lakers of']
    mock_aioresponse.assert_called_once_with(url, method='POST',
                                             headers={AUTH_HEADER: BEARER_TOKEN + llm_client._api_key},
                                             json={'inputs': 'who is kobe bryant',"max_length": 10, "temperature": 1.0, "top_p" : None},
                                             raise_for_status=True)


@pytest.mark.asyncio
def test_get_tokens_count__sanity(mock_aioresponse, llm_client, url):
    actual = llm_client.get_tokens_count(text="is queen elisabeth alive?")
    assert actual == 7



import pytest

from sfn_llm_client.llm_api_client.huggingface_client import HuggingFaceClient, BASE_URL
from sfn_llm_client.llm_api_client.base_llm_api_client import LLMAPIClientConfig


@pytest.fixture
def model_name():
    return "oasst-sft-4-pythia-12b-epoch-3.5"


@pytest.fixture
def config(client_session, model_name):
    return LLMAPIClientConfig("top-secret-api-key", client_session, default_model=model_name)


@pytest.fixture
def llm_client(config):
    return HuggingFaceClient(config)


@pytest.fixture
def url(model_name):
    return build_url(model_name)


def build_url(model: str) -> str:
    return BASE_URL + model + "/"



from unittest.mock import patch

import pytest

from sfn_llm_client import LLMAPIClientType, LLMAPIClientFactory, LLMAPIClientConfig
from sfn_llm_client.llm_api_client.llm_api_client_factory import get_llm_api_client_class


def test_get_llm_api_client__without_context_manager():
    llm_api_client_factory = LLMAPIClientFactory()
    with pytest.raises(ValueError):
        llm_api_client_factory.get_llm_api_client(LLMAPIClientType.OPEN_AI, api_key="super secret key")


@pytest.mark.asyncio
@pytest.mark.parametrize("client_type,client_patch",
                         [(LLMAPIClientType.OPEN_AI, "OpenAIClient"), (LLMAPIClientType.AI21, "AI21Client"),
                          (LLMAPIClientType.HUGGING_FACE, "HuggingFaceClient"),
                          (LLMAPIClientType.ALEPH_ALPHA, "AlephAlphaClient"),
                          (LLMAPIClientType.ANTHROPIC, "AnthropicClient"),
                          (LLMAPIClientType.GOOGLE, "GoogleClient")])
async def test_get_llm_api_client__with_client_type(client_type, client_patch):
    assert len(LLMAPIClientType) == 6

    llm_api_client_factory = LLMAPIClientFactory()
    async with llm_api_client_factory:
        with patch(f"sfn_llm_client.{client_patch}") as mock_client:
            actual = llm_api_client_factory.get_llm_api_client(client_type, api_key="super secret key")

            assert actual is mock_client.return_value
            mock_client.assert_called_once_with(LLMAPIClientConfig(session=llm_api_client_factory._session,
                                                                   api_key="super secret key"))


@pytest.mark.asyncio
async def test_get_llm_api_client__with_unknown_client_type():
    llm_api_client_factory = LLMAPIClientFactory()
    async with llm_api_client_factory:
        with pytest.raises(ValueError):
            llm_api_client_factory.get_llm_api_client("unknown-client-type", api_key="super secret key")


@pytest.mark.parametrize("client_type,client_patch",
                         [(LLMAPIClientType.OPEN_AI, "OpenAIClient"), (LLMAPIClientType.AI21, "AI21Client"),
                          (LLMAPIClientType.HUGGING_FACE, "HuggingFaceClient"),
                          (LLMAPIClientType.ALEPH_ALPHA, "AlephAlphaClient"),
                          (LLMAPIClientType.ANTHROPIC, "AnthropicClient"),
                          (LLMAPIClientType.GOOGLE, "GoogleClient")])
def test_get_llm_api_client_class(client_type, client_patch):
    assert len(LLMAPIClientType) == 6

    with patch(f"sfn_llm_client.{client_patch}") as mock_client:
        actual = get_llm_api_client_class(client_type)

        assert actual is mock_client
        mock_client.assert_not_called()


def test_get_llm_api_client_class__with_unknown_client_type():
    with pytest.raises(ValueError):
        get_llm_api_client_class("unknown-client-type")



import logging
import threading
from uuid import UUID
from typing import Any, Dict, List, Optional

from langchain_core.outputs import LLMResult
from langchain_core.callbacks import BaseCallbackHandler

from sfn_llm_client.llm_api_client.core.model_schema import MODEL_COST_PER_1M_TOKENS, PROVIDER_TO_BASE_CLASS, Provider

class CostCallbackHandler(BaseCallbackHandler):
    total_tokens: int = 0
    prompt_tokens: int = 0
    completion_tokens: int = 0
    guardrails_tokens: int = 0
    successful_requests: int = 0
    total_cost: float = 0.0

    def __init__(
        self,
        logger: logging.Logger,
        cost_dict: Dict = MODEL_COST_PER_1M_TOKENS,
        provider_map: Dict = PROVIDER_TO_BASE_CLASS,
    ) -> None:
        super().__init__()
        self.logger = logger
        self._lock = threading.Lock()
        self.cost_dict = cost_dict
        self._base_class_to_provider = {v: k for k, v in provider_map.items()}
        self._run_info: Dict[UUID, Dict[str, Any]] = {}

    def __repr__(self) -> str:
        return (
            f"Tokens Used: {self.total_tokens}\n"
            f"\tPrompt Tokens: {self.prompt_tokens}\n"
            f"\tCompletion Tokens: {self.completion_tokens}\n"
            f"\tGuardrails Tokens: {self.guardrails_tokens}\n"
            f"Successful Requests: {self.successful_requests}\n"
            f"Total Cost (USD): ${self.total_cost:.6f}"
        )

    def on_llm_start(
        self, serialized: Dict[str, Any], prompts: List[str], *, run_id: UUID, **kwargs: Any
    ) -> None:
        class_hierarchy = serialized.get("id", [])
        model_provider = next(
            (self._base_class_to_provider[cls] for cls in class_hierarchy if cls in self._base_class_to_provider),
            None,
        )
        model_id = serialized.get("kwargs", {}).get("model_name")

        if model_provider == Provider.SNOWFLAKE:
            model_id = kwargs.get("metadata", {}).get("ls_model_name", None)            

        if not model_provider or not model_id:
            self.logger.warning(
                f"Could not determine provider or model for run {run_id}. "
                f"Cost will not be calculated. Class Hierarchy: {class_hierarchy}, Model ID: {model_id}"
            )
            return

        if model_provider == Provider.FAKE:
            self.logger.debug(f"Skipping cost calculation for Fake model run {run_id}.")
            return

        with self._lock:
            self._run_info[run_id] = {"provider": model_provider, "model_name": model_id}

    def on_llm_end(self, response: LLMResult, *, run_id: UUID, **kwargs: Any) -> None:
        with self._lock:
            run_info = self._run_info.pop(run_id, None)

        if not run_info:
            if self.logger.isEnabledFor(logging.DEBUG):
                self.logger.debug(f"No start information for run {run_id}. This is normal for fake models.")
            return
        
        provider: Provider = run_info["provider"]
        model_name: str = run_info["model_name"]
        
        if provider == Provider.SNOWFLAKE:
            generations = getattr(response, 'generations', [])
            generation = (generations[0] if generations else [{}])[0]
            message = getattr(generation, 'message', None)
            token_usage = getattr(message, 'response_metadata', {}) 

        else:
            llm_output = response.llm_output or {}
            token_usage = llm_output.get("token_usage")

        if not token_usage:
            self.logger.warning(f"No token_usage information in LLM output for run {run_id}. Cost cannot be calculated.")
            return

        
        prompt_tokens = token_usage.get("prompt_tokens", 0)
        completion_tokens = token_usage.get("completion_tokens", 0)
        guardrails_tokens = token_usage.get("guardrails_tokens", 0) if provider == Provider.SNOWFLAKE else 0
        
        run_cost = 0.0
        model_costs = self.cost_dict.get(provider.value, {}).get(model_name)

        if model_costs:
            if provider == Provider.SNOWFLAKE:
                total_run_tokens = prompt_tokens + completion_tokens + guardrails_tokens
                cost_per_mil = model_costs.get("cost", 0.0)
                run_cost = (total_run_tokens / 1_000_000) * cost_per_mil
            else:
                input_cost_per_mil = model_costs.get("input_cost", 0.0)
                output_cost_per_mil = model_costs.get("output_cost", 0.0)
                run_cost = (prompt_tokens * input_cost_per_mil + completion_tokens * output_cost_per_mil) / 1_000_000
        else:
             self.logger.warning(
                f"Cost not found for model '{model_name}' from provider '{provider.value}'. "
                f"Cost will be zero for this run."
            )

        with self._lock:
            self.total_tokens += prompt_tokens + completion_tokens + guardrails_tokens
            self.prompt_tokens += prompt_tokens
            self.completion_tokens += completion_tokens
            self.guardrails_tokens += guardrails_tokens
            self.total_cost += run_cost
            self.successful_requests += 1
    
    def reset(self) -> None:
        with self._lock:
            self.total_tokens = 0
            self.prompt_tokens = 0
            self.completion_tokens = 0
            self.guardrails_tokens = 0
            self.successful_requests = 0
            self.total_cost = 0.0
            self.logger.info("Cost tracker has been reset.")

    def __copy__(self) -> "CostCallbackHandler":
        return self

    def __deepcopy__(self, memo: Any) -> "CostCallbackHandler":
        return self


from typing import Optional

from sfn_llm_client.llm_api_client.base_llm_api_client import BaseLLMAPIClient, LLMAPIClientConfig
from sfn_llm_client.utils.consts import PROMPT_KEY

COMPLETE_PATH = "complete"
TOKENIZE_PATH = "tokenize"
BASE_URL = "https://api.ai21.com/studio/v1/"
COMPLETIONS_KEY = "completions"
DATA_KEY = "data"
TEXT_KEY = "text"
TOKENS_KEY = "tokens"
AUTH_HEADER = "Authorization"
BEARER_TOKEN = "Bearer "


class AI21Client(BaseLLMAPIClient):
    def __init__(self, config: LLMAPIClientConfig):
        super().__init__(config)
        if self._base_url is None:
            self._base_url = BASE_URL
        self._headers[AUTH_HEADER] = BEARER_TOKEN + self._api_key

    async def text_completion(self, prompt: str, model: Optional[str] = None, max_tokens: int = 16,
                              temperature: float = 0.7, top_p: float = 1,**kwargs) -> list[str]:
        model = model or self._default_model
        kwargs[PROMPT_KEY] = prompt
        kwargs["topP"] = kwargs.pop("topP", top_p)
        kwargs["maxTokens"] = kwargs.pop("maxTokens", max_tokens)
        kwargs["temperature"] = temperature
        response = await self._session.post(self._base_url + model + "/" + COMPLETE_PATH,
                                            json=kwargs,
                                            headers=self._headers,
                                            raise_for_status=True)
        response_json = await response.json()
        completions = response_json[COMPLETIONS_KEY]
        return [completion[DATA_KEY][TEXT_KEY] for completion in completions]

    async def get_tokens_count(self, text: str, **kwargs) -> int:
        response = await self._session.post(self._base_url + TOKENIZE_PATH,
                                            json={TEXT_KEY: text},
                                            headers=self._headers,
                                            raise_for_status=True)
        response_json = await response.json()
        return len(response_json[TOKENS_KEY])



from setuptools import setup, find_packages

# To Load the README file to use as the long description
with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="sfn_llm_client",
    # version="0.1.2",  # You can dynamically fetch the version if needed
    description="SDK for using LLM clients",
    long_description=long_description,  # Using README.md as long description
    long_description_content_type="text/markdown",  # README file format
    author="Rajesh Darak",
    author_email="rajesh@stepfuction.ai",
    url="https://github.com/iamrajeshdaraksfn/llm-client-sdk.git",
    classifiers=[
        "Operating System :: OS Independent",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
    python_requires='>=3.9',
    packages=find_packages(),
    install_requires=[
        "aiohttp >=3.0.0,<4.0.0",
        "dataclasses_json >= 0.5.0",
        "openai >=1.54.3",
        "tiktoken >=0.3.3",
        "anthropic >= 0.39.0",
        "snowflake-connector-python >=3.12.3",
        "snowflake-ml-python==1.7.0",
        "snowflake-snowpark-python==1.23.0",
        "transformers >= 4.46.2",
        "langchain-openai >= 0.3.28",
        "snowflake-snowpark-python >=1.0.0",
        "langchain-community >= 0.0.30",
        "pydantic >=2.6.0, <3.0.0",
        "langchain-core>=0.3,<0.4",
        "StrEnum>=0.4.14"
    ],
    extras_require={
        'test': [
            "pytest",
            "pytest-aiohttp",
            "pytest-asyncio",
            "pytest-mock",
            "aioresponses",
        ],
        'openai': [
            "openai >=1.54.3",
            "tiktoken >=0.3.3",
        ],
        'huggingface': [
            "transformers >= 4.0.0",
        ],
        'anthropic': [
            "anthropic >= 0.39.0",
        ],
        'arctic': [
            "snowflake-connector-python >=3.12.3",
            "snowflake-ml-python==1.7.0",
            "snowflake-snowpark-python==1.23.0",
            "transformers >= 4.46.2",
        ],
        'openai_langchain': [
            "langchain-openai >= 0.3.28",
        ],
        'cortex_langchain': [
            "snowflake-snowpark-python >=1.0.0",
            "langchain-community >= 0.0.30",
            "pydantic >=2.6.0, <3.0.0",
            "langchain-core>=0.3,<0.4",
        ],
        'google': [
            "google-generativeai >= 0.1.0",
        ],
        'api': [
            "sfn_llm_client[openai,huggingface,anthropic,google,arctic]",
        ],
        'local': [
            "transformers >= 4.0.0",
        ],
        'sync': [
            "async_to_sync >= 0.2.0",
        ],
        'all': [
            "sfn_llm_client[api,local,sync]",
        ],
    },
    package_data={
        "sfn_llm_client": ["README.md"],  # Any other non-Python files can be listed here
    },
)



from importlib import util
from langchain_community.chat_models import FakeListChatModel
from langchain_core.language_models import BaseChatModel
from pydantic import model_validator
from typing import Dict, Optional
from functools import cache
from langchain_core.utils.input import get_colored_text, get_bolded_text
from langchain_core.tracers.stdout import try_json_stringify


from sfn_llm_client.llm_api_client.core.model_schema import LLMConfig, Provider
from sfn_llm_client.utils.logging import setup_logger


# class FakeToolModel(FakeListChatModel):
#     def __init__(self, responses: list[str]):
#         super().__init__(responses=responses)
#     def bind_tools(self, tools):
#         return self


logger, _ = setup_logger(logger_name="LLMClient")

def _check_pkg(pkg: str, *, pkg_kebab: Optional[str] = None) -> None:
    if not util.find_spec(pkg):
        pkg_kebab = pkg_kebab if pkg_kebab is not None else pkg.replace("_", "-")
        msg = (
            f"Unable to import {pkg}. Please install with `pip install -U {pkg_kebab}`"
        )
        logger.error(msg)
        raise ImportError(msg)

def load_model( config: LLMConfig) -> BaseChatModel:

    provider, model = config.model_name.split("/", maxsplit=1)
    temperature=config.temperature
    top_p=config.top_p
    max_retries=config.max_retries
    timeout=config.api_timeout
    cortex_function=config.cortex_function

    if provider == Provider.OPENAI.value:
        _check_pkg("langchain_openai")
        from langchain_openai import ChatOpenAI
        logger.info("Creating OpenAI model instance: '{}'".format(model))
        return ChatOpenAI(
            model=model,
            temperature=temperature,
            top_p=top_p,
            max_retries=max_retries,
            timeout=timeout
        )
    if provider == Provider.SNOWFLAKE.value:
        _check_pkg("langchain_community")
        _check_pkg("snowflake.snowpark", pkg_kebab="snowflake-snowpark-python")
        from sfn_llm_client.llm_api_client.core.custom_snowflake import CustomChatSnowflakeCortex
        logger.info("Creating Snowflake model instance: '{}'".format(model))
        return CustomChatSnowflakeCortex(
            model=model, 
            temperature=temperature,
            top_p=top_p,
            cortex_function=cortex_function,
        )
    
    msg = (
        f"Unsupported {provider}"
    )
    logger.error(msg)
    raise ValueError(msg)

    


# def _create_single_model(config: LLMConfig) -> BaseChatModel:

#     provider, model_id = config.model_name.split("/", maxsplit=1)

#     if provider == Provider.OPENAI.value:
#         if ChatOpenAI is None:
#             config.logger.error("OpenAI dependencies not found. Run 'pip install langchain-openai'")
#             raise ImportError("OpenAI dependencies not found. Run 'pip install langchain-openai'.")
        
#         config.logger.info(f"Creating OpenAI model instance: '{model_id}'")
#         return ChatOpenAI(
#             model=model_id,
#             temperature=config.temperature,
#             top_p=config.top_p,
#             max_retries=config.max_retries,
#             timeout=config.api_timeout
#         )

#     elif provider == Provider.SNOWFLAKE.value:
#         if CustomChatSnowflakeCortex is None:
#             config.logger.error("Snowflake dependencies not found. Run 'pip install langchain-community snowflake-snowpark-python'.")
#             raise ImportError("Snowflake dependencies not found. Run 'pip install langchain-community snowflake-snowpark-python'.")
        
#         config.logger.info(f"Creating Snowflake Cortex model instance: '{model_id}'")
#         return CustomChatSnowflakeCortex(
#                 model=model_id, 
#                 temperature=config.temperature,
#                 top_p=config.top_p,
#                 cortex_function=config.cortex_function,
#             )

#     elif provider == Provider.FAKE.value:
#         config.logger.info(f"Creating Fake model instance: '{model_id}'")
#         responses = config.fake_responses or ["This is a default fake response."]
#         return FakeToolModel(responses=responses)

#     config.logger.error(f"Unsupported provider: '{provider}'. This should have been caught by validation.")
#     raise ValueError(f"Unsupported provider: {provider}")




# def get_model( config: LLMConfig) -> BaseChatModel:
#     if not config:
#         temp_logger = logging.getLogger(__name__)
#         temp_logger.error("Model configuration is required but was not provided.")
#         raise ValueError("Model configuration is required")

#     primary_llm = _create_single_model(config)
#     if not config.fall_back_model:
#         config.logger.info(f"Model '{config.model_name}' loaded without a fallback.")
#         return primary_llm

#     config.logger.info(f"Preparing fallback model '{config.fall_back_model}' for primary model '{config.model_name}'.")

#     fallback_config = LLMConfig(
#         model_name=config.fall_back_model,
#         fall_back_model=config.fall_back_model, 
#         temperature=config.temperature,
#         top_p=config.top_p,
#         max_retries=2,
#         api_timeout=config.api_timeout,
#         logger=config.logger,
#     )
#     fallback_llm = _create_single_model(fallback_config)
#     return primary_llm.with_fallbacks([fallback_llm])






    
    


from unittest.mock import AsyncMock, MagicMock

import pytest
from aiohttp import ClientSession
from openai.openai_object import OpenAIObject

from sfn_llm_client import OpenAIClient, LLMAPIClientType, LLMAPIClientFactory
from sfn_llm_client.llm_api_client.base_llm_api_client import LLMAPIClientConfig, Role
from sfn_llm_client.llm_api_client.openai_client import ChatMessage
from tests.test_utils.load_json_resource import load_json_resource


@pytest.mark.asyncio
async def test_get_llm_api_client__with_open_ai(config):
    del config.session
    async with LLMAPIClientFactory() as llm_api_client_factory:
        actual = llm_api_client_factory.get_llm_api_client(LLMAPIClientType.OPEN_AI, **config.__dict__)

    assert isinstance(actual, OpenAIClient)


def test_init__sanity(openai_mock, client_session):
    OpenAIClient(LLMAPIClientConfig("fake_api_key", client_session))

    assert openai_mock.api_key == "fake_api_key"
    openai_mock.aiosession.set.assert_called_once()


@pytest.mark.asyncio
async def test_text_completion__sanity(openai_mock, open_ai_client, model_name):
    openai_mock.Completion.acreate = AsyncMock(
        return_value=OpenAIObject.construct_from(load_json_resource("openai/text_completion.json")))
    actual = await open_ai_client.text_completion(prompt="These are a few of my favorite")

    assert actual == ["\n\nThis is indeed a test"]
    openai_mock.Completion.acreate.assert_awaited_once_with(
        model=model_name,
        prompt="These are a few of my favorite",
        headers={}, temperature=0, max_tokens=16, top_p=1)


@pytest.mark.asyncio
async def test_text_completion__return_multiple_completions(openai_mock, open_ai_client, model_name):
    open_ai_object = OpenAIObject.construct_from(load_json_resource("openai/text_completion.json"))
    open_ai_object.choices.append(OpenAIObject.construct_from({"text": "second completion"}))
    openai_mock.Completion.acreate = AsyncMock(return_value=open_ai_object)

    actual = await open_ai_client.text_completion(prompt="These are a few of my favorite")

    assert actual == ["\n\nThis is indeed a test", "second completion"]
    openai_mock.Completion.acreate.assert_awaited_once_with(
        model=model_name,
        prompt="These are a few of my favorite",
        headers={}, temperature=0, max_tokens=16, top_p=1)


@pytest.mark.asyncio
async def test_text_completion__override_model(openai_mock, open_ai_client, model_name):
    new_model_name = "gpt3"
    openai_mock.Completion.acreate = AsyncMock(
        return_value=OpenAIObject.construct_from(load_json_resource("openai/text_completion.json")))

    actual = await open_ai_client.text_completion(prompt="These are a few of my favorite", model=new_model_name)

    assert actual == ["\n\nThis is indeed a test"]
    openai_mock.Completion.acreate.assert_awaited_once_with(
        model=new_model_name,
        prompt="These are a few of my favorite",
        headers={}, temperature=0, max_tokens=16, top_p=1)


@pytest.mark.asyncio
async def test_text_completion__with_kwargs(openai_mock, open_ai_client, model_name):
    openai_mock.Completion.acreate = AsyncMock(
        return_value=OpenAIObject.construct_from(load_json_resource("openai/text_completion.json")))

    actual = await open_ai_client.text_completion(prompt="These are a few of my favorite", max_tokens=10)

    assert actual == ["\n\nThis is indeed a test"]
    openai_mock.Completion.acreate.assert_awaited_once_with(
        model=model_name,
        prompt="These are a few of my favorite",
        temperature=0, max_tokens=10, top_p=1,
        headers={})


@pytest.mark.asyncio
async def test_text_completion__with_headers(openai_mock, model_name):
    openai_mock.Completion.acreate = AsyncMock(
        return_value=OpenAIObject.construct_from(load_json_resource("openai/text_completion.json")))
    open_ai_client = OpenAIClient(LLMAPIClientConfig("fake_api_key", MagicMock(ClientSession), default_model=model_name,
                                                     headers={"header_name": "header_value"}))

    actual = await open_ai_client.text_completion(prompt="These are a few of my favorite")

    assert actual == ["\n\nThis is indeed a test"]
    openai_mock.Completion.acreate.assert_awaited_once_with(
        model=model_name,
        prompt="These are a few of my favorite",
        headers={"header_name": "header_value"}, temperature=0, max_tokens=16, top_p=1)


@pytest.mark.asyncio
async def test_chat_completion__sanity(openai_mock, open_ai_client, model_name):
    openai_mock.ChatCompletion.acreate = AsyncMock(
        return_value=OpenAIObject.construct_from(load_json_resource("openai/chat_completion.json")))

    actual = await open_ai_client.chat_completion([ChatMessage(Role.USER, "Hello!")])

    assert actual == ["\n\nHello there, how may I assist you today?"]
    openai_mock.ChatCompletion.acreate.assert_awaited_once_with(
        model=model_name,
        messages=[{'content': 'Hello!', 'role': 'user'}],
        headers={}, temperature=0, max_tokens=16, top_p=1)


@pytest.mark.asyncio
async def test_chat_completion__return_multiple_completions(openai_mock, open_ai_client, model_name):
    open_ai_object = OpenAIObject.construct_from(load_json_resource("openai/chat_completion.json"))
    open_ai_object.choices.append(OpenAIObject.construct_from({"message": {"content": "second completion"}}))
    openai_mock.ChatCompletion.acreate = AsyncMock(return_value=open_ai_object)

    actual = await open_ai_client.chat_completion([ChatMessage(Role.USER, "Hello!")])

    assert actual == ["\n\nHello there, how may I assist you today?", "second completion"]
    openai_mock.ChatCompletion.acreate.assert_awaited_once_with(
        model=model_name,
        messages=[{'content': 'Hello!', 'role': 'user'}],
        headers={}, temperature=0, max_tokens=16, top_p=1)


@pytest.mark.asyncio
async def test_chat_completion__override_model(openai_mock, open_ai_client, model_name):
    new_model_name = "gpt3"
    openai_mock.ChatCompletion.acreate = AsyncMock(
        return_value=OpenAIObject.construct_from(load_json_resource("openai/chat_completion.json")))

    actual = await open_ai_client.chat_completion([ChatMessage(Role.USER, "Hello!")], model=new_model_name)

    assert actual == ["\n\nHello there, how may I assist you today?"]
    openai_mock.ChatCompletion.acreate.assert_awaited_once_with(
        model=new_model_name,
        messages=[{'content': 'Hello!', 'role': 'user'}],
        headers={}, temperature=0, max_tokens=16, top_p=1)


@pytest.mark.asyncio
async def test_chat_completion__with_kwargs(openai_mock, open_ai_client, model_name):
    openai_mock.ChatCompletion.acreate = AsyncMock(
        return_value=OpenAIObject.construct_from(load_json_resource("openai/chat_completion.json")))

    actual = await open_ai_client.chat_completion([ChatMessage(Role.USER, "Hello!")], max_tokens=10, top_p=1)

    assert actual == ["\n\nHello there, how may I assist you today?"]
    openai_mock.ChatCompletion.acreate.assert_awaited_once_with(
        model=model_name,
        messages=[{'content': 'Hello!', 'role': 'user'}],
        max_tokens=10,
        headers={}, temperature=0, top_p=1)


@pytest.mark.asyncio
async def test_chat_completion__with_headers(openai_mock, model_name):
    openai_mock.ChatCompletion.acreate = AsyncMock(
        return_value=OpenAIObject.construct_from(load_json_resource("openai/chat_completion.json")))
    open_ai_client = OpenAIClient(LLMAPIClientConfig("fake_api_key", MagicMock(ClientSession), default_model=model_name,
                                                     headers={"header_name": "header_value"}))

    actual = await open_ai_client.chat_completion([ChatMessage(Role.USER, "Hello!")])

    assert actual == ["\n\nHello there, how may I assist you today?"]
    openai_mock.ChatCompletion.acreate.assert_awaited_once_with(
        model=model_name,
        messages=[{'content': 'Hello!', 'role': 'user'}],
        headers={"header_name": "header_value"}, temperature=0, max_tokens=16, top_p=1)


@pytest.mark.asyncio
@pytest.mark.parametrize("model_name,expected", [("gpt-3.5-turbo-0301", 127), ("gpt-3.5-turbo-0613", 129),
                                                 ("gpt-3.5-turbo", 129), ("gpt-4-0314", 129), ("gpt-4-0613", 129),
                                                 ("gpt-4", 129)])
async def test_get_chat_tokens_count__with_examples_from_openai_cookbook(model_name, expected, open_ai_client):
    example_messages = [
        ChatMessage(Role.SYSTEM,
                    "You are a helpful, pattern-following assistant that translates corporate jargon "
                    "into plain English."),
        ChatMessage(Role.SYSTEM, "New synergies will help drive top-line growth.", name="example_user"),
        ChatMessage(Role.SYSTEM, "Things working well together will increase revenue.", name="example_assistant"),
        ChatMessage(Role.SYSTEM,
                    "Let's circle back when we have more bandwidth to touch base on opportunities "
                    "for increased leverage.", name="example_user"),
        ChatMessage(Role.SYSTEM, "Let's talk later when we're less busy about how to do better.",
                    name="example_assistant"),
        ChatMessage(Role.USER,
                    "This late pivot means we don't have time to boil the ocean for the client deliverable."),
    ]

    actual = await open_ai_client.get_chat_tokens_count(example_messages, model=model_name)

    assert actual == expected


@pytest.mark.asyncio
async def test_get_tokens_count__sanity(model_name, open_ai_client, tiktoken_mock):
    tokeniser_mock = tiktoken_mock.encoding_for_model.return_value
    tokeniser_mock.encode.return_value = [123, 456]
    text = "This is a test"

    actual = await open_ai_client.get_tokens_count(text=text)

    assert actual == len(tokeniser_mock.encode.return_value)
    tiktoken_mock.encoding_for_model.assert_called_once_with(model_name)
    tokeniser_mock.encode.assert_called_once_with(text)


@pytest.mark.asyncio
async def test_get_tokens_count__override_model(open_ai_client, tiktoken_mock):
    tokeniser_mock = tiktoken_mock.encoding_for_model.return_value
    tokeniser_mock.encode.return_value = [123, 456]
    text = "This is a test"
    model_name = "gpt3"

    actual = await open_ai_client.get_tokens_count(text=text, model=model_name)

    assert actual == len(tokeniser_mock.encode.return_value)
    tiktoken_mock.encoding_for_model.assert_called_once_with(model_name)
    tokeniser_mock.encode.assert_called_once_with(text)



from typing import Optional
import time
from anthropic import Anthropic

from sfn_llm_client.llm_api_client.base_llm_api_client import BaseLLMAPIClient, LLMAPIClientConfig, ChatMessage, Role
from sfn_llm_client.utils.consts import PROMPT_KEY
from sfn_llm_client.utils.logging import setup_logger
from sfn_llm_client.utils.retry_with import retry_with
from sfn_llm_client.llm_cost_calculation.anthropic_cost_calculation import anthropic_cost_calculation

COMPLETE_PATH = "complete"
BASE_URL = "https://api.anthropic.com/v1/"
COMPLETIONS_KEY = "completion"
AUTH_HEADER = "x-api-key"
ACCEPT_HEADER = "Accept"
VERSION_HEADER = "anthropic-version"
ACCEPT_VALUE = "application/json"
MAX_TOKENS_KEY = "max_tokens_to_sample"
USER_PREFIX = "Human:"
ASSISTANT_PREFIX = "Assistant:"
START_PREFIX = "\n\n"
SYSTEM_START_PREFIX = "<admin>"
SYSTEM_END_PREFIX = "</admin>"


class AnthropicClient(BaseLLMAPIClient):
    def __init__(self, config: LLMAPIClientConfig):
        super().__init__(config)
        self._base_url = config.base_url or BASE_URL
        self._anthropic = Anthropic(api_key=config.api_key)
        self.logger, _ = setup_logger(logger_name="OpenAIClient")
        self._headers = {
            VERSION_HEADER: self._anthropic.default_headers[VERSION_HEADER],
            ACCEPT_HEADER: ACCEPT_VALUE,
            AUTH_HEADER: config.api_key,
        }

    @retry_with(retries=3, retry_delay=3.0, backoff=True)
    def chat_completion(self, messages: list[ChatMessage], model: Optional[str] = None,
                    max_tokens: Optional[int] = None, temperature: float = 1.0, retries: int = 3,
                    retry_delay: float = 3.0, **kwargs) -> list[str]:
        """
        This method performs chat completion with retry logic for handling exceptions or empty responses.
        It also calculates token consumption based on the API response.

        :param messages: List of ChatMessage objects representing the conversation history.
        :param model: Optional model name to be used.
        :param max_tokens: Maximum number of tokens to generate in the response.
        :param temperature: Controls the creativity of the response.
        :param retries: Number of retries in case of failure.
        :param retry_delay: Delay in seconds between retries.

        :return: ChatCompletion object containing the generated text and other metadata,
                or None if all retries fail.
        """

        self.logger.info("Started running llm client sdk chat completion...")
        del kwargs['session'] #deleting session key not required for anthropic
        response = self._anthropic.messages.create(model=model, max_tokens=max_tokens, messages=messages, **kwargs)

        # Calculate token consumption cost
        token_cost_summary = anthropic_cost_calculation(
            response.usage.input_tokens,
            response.usage.output_tokens,
            model=model,
        )

        return response, token_cost_summary


    def text_completion(self, prompt: str, model: Optional[str] = None, max_tokens: Optional[int] = None,
                         temperature: float = 1.0, top_p: Optional[float] = None, **kwargs) -> list[str]:
        if max_tokens is None and kwargs.get(MAX_TOKENS_KEY) is None:
            raise ValueError(f"max_tokens or {MAX_TOKENS_KEY} must be specified")

        kwargs[PROMPT_KEY] = prompt
        kwargs[MAX_TOKENS_KEY] = kwargs.pop(MAX_TOKENS_KEY, max_tokens)
        kwargs["temperature"] = temperature
        if top_p:
            kwargs["top_p"] = top_p

        self._set_model_in_kwargs(kwargs, model)
        response = self._anthropic.messages.create(model=model, messages=[{"content": prompt}], **kwargs)
        return [response[COMPLETIONS_KEY]]

    def get_chat_tokens_count(self, messages: list[ChatMessage], **kwargs) -> int:
        return self.get_tokens_count(self.messages_to_text(messages), **kwargs)

    def get_tokens_count(self, text: str, **kwargs) -> int:
        return self._anthropic.count_tokens(text)
        # return sum(len(word.split()) for word in text.split("\n"))  # Approximate token count based on words

    def messages_to_text(self, messages: list[ChatMessage]) -> str:
        prompt = START_PREFIX
        prompt += START_PREFIX.join(map(self._message_to_prompt, messages))
        if messages[-1].role != Role.ASSISTANT:
            prompt += START_PREFIX
            prompt += self._message_to_prompt(ChatMessage(role=Role.ASSISTANT, content=""))
        return prompt.rstrip()

    @staticmethod
    def _message_to_prompt(message: ChatMessage) -> str:
        if message.role == Role.USER:
            return f"{USER_PREFIX} {message.content}"
        elif message.role == Role.ASSISTANT:
            return f"{ASSISTANT_PREFIX} {message.content}"
        elif message.role == Role.SYSTEM:
            return f"{USER_PREFIX} {SYSTEM_START_PREFIX}{message.content}{SYSTEM_END_PREFIX}"
        else:
            raise ValueError(f"Unknown role: {message.role}")